from pathlib import Path
import sys
import os

from sailreval.utils.binary_debug_info import read_line_maps, gen_dwarf_addr_to_line_map, dump_dwarf_addr_to_line_map
from sailreval.metrics.ged_to_source import compute_cfg_edit_distance
from sailreval.decompilers.angr_dec import angr_decompile
from sailreval.decompilers.ida_dec import ida_decompile
from sailreval import SAILR_DECOMPILERS

import unittest
from pyjoern import fast_cfgs_from_source
from pyjoern.mapping import correct_source_cfg_addrs


FILES_PATH = Path(os.path.join(os.path.dirname(os.path.realpath(__file__)), "./cfged/"))
DECOMPILERS = {
    SAILR_DECOMPILERS.SOURCE_CODE: None,
    SAILR_DECOMPILERS.ANGR_SAILR: angr_decompile,
    SAILR_DECOMPILERS.IDA: ida_decompile
}

#
# helper functions
#


def extract_cfgs_for_decompiler(source_dir: Path, binary_name, decompiler, function=None):
    source_dir = Path(source_dir).absolute()
    out = fast_cfgs_from_source(source_dir.joinpath(f"{decompiler}_{binary_name}.c"))

    if decompiler == SAILR_DECOMPILERS.SOURCE_CODE:
        linemaps = source_dir.joinpath(f"source_{binary_name}.linemaps")
        out = correct_source_cfg_addrs(out, linemaps)

    return out[function] if function else out


def angr_decompile_and_get_dwarf_info(binary_path: Path, functions=None):
    angr_decompile(binary_path, functions=functions, print_dec=False, dump_line_maps=True)
    dump_dwarf_addr_to_line_map(binary_path)


def cleanup_angr_and_src_files(binary_path: Path, binary_name: str):
    binary_path = Path(binary_path).absolute()
    binary_path.parent.joinpath(f"angr_sailr_{binary_name}.linemaps").unlink()
    binary_path.parent.joinpath(f"angr_sailr_{binary_name}.c").unlink()
    binary_path.parent.joinpath(f"{binary_name}.dwarf.linemaps").unlink()


def angr_dec_cfged_score(binary_path: Path, functions=None):
    filename = binary_path.with_suffix("").name
    decompiler = SAILR_DECOMPILERS.ANGR_SAILR
    angr_decompile_and_get_dwarf_info(binary_path, functions=functions)
    scores = {}
    for function in functions:
        src_cfg = extract_cfgs_for_decompiler(binary_path.parent, filename, "source", function=function)
        dec_cfg = extract_cfgs_for_decompiler(binary_path.parent, filename, decompiler, function=function)
        cfged_score = compute_cfg_edit_distance(dec_cfg, src_cfg, function, binary_path, decompiler)
        scores[function] = cfged_score

    cleanup_angr_and_src_files(binary_path, filename)
    return scores

#
# tests
#


class TestReadability(unittest.TestCase):
    def test_shred_dosync(self):
        cfged_scores = angr_dec_cfged_score(FILES_PATH / "shred.o", functions=["dosync"])
        self.assertEqual(cfged_scores["dosync"], 0)

    def test_tsort_record_relation(self):
        cfged_scores = angr_dec_cfged_score(FILES_PATH / "tsort.o", functions=["record_relation"])
        self.assertEqual(cfged_scores["record_relation"], 0)

    def test_ptx_sort_found_occurs(self):
        cfged_scores = angr_dec_cfged_score(FILES_PATH / "ptx.o", functions=["sort_found_occurs"])
        self.assertEqual(cfged_scores["sort_found_occurs"], 0)


if __name__ == "__main__":
    unittest.main(argv=sys.argv)



__version__ = "1.4.0"

# create loggers
import logging
logging.getLogger("sailreval").addHandler(logging.NullHandler())
from sailreval.loggercfg import Loggers
loggers = Loggers()
del Loggers

l = logging.getLogger(__name__)


class SAILR_METRICS:
    CODE_COMPLEXITY = "cc"
    GOTO_COUNT = "gotos"
    LINES_OF_CODE = "loc"
    IF_STMT_COUNT = "ifs"
    FUNC_CALLS = "func_calls"
    PATH_COUNT = "paths"
    BOOLEAN_COUNT = "bools"
    GED_UPPERBOUND = "ged_upperbound"
    GED_MAX = "ged_max"
    GED_EXACT = "ged_exact"
    CFGED = "cfged"
    GRAPH_SIZE = "graph_size"
    BLOCK_COUNT = "block_count"
    DEC_TIME = "dec_time"

    # Post Metrics
    NORM_CFGED = "norm_cfged"               # CFGED / GRAPH_SZE
    ZERO_CFGED = "zero_cfged"               # count of CFGED == 0
    FUNCS_W_GOTO = "funcs_w_goto"           # count of goto >= 1
    GOTO_FUNC_CFGED = "goto_func_cfged"     # CFGED of functions with goto >= 1



class SAILR_DECOMPILERS:
    SOURCE_CODE = "source"
    SOURCE_EXPANDED_CODE = "expanded_source"
    ANGR_SAILR = "angr_sailr"
    ANGR_IMPROVED_PHOENIX = "angr_improved_phoenix"
    ANGR_PHOENIX = "angr_phoenix"
    ANGR_DREAM = "angr_dream"
    ANGR_COMB = "angr_comb"
    IDA = "ida"
    GHIDRA = "ghidra"


ALL_METRICS = [getattr(SAILR_METRICS, a) for a in dir(SAILR_METRICS) if not a.startswith("_")]
ALL_DECOMPILERS = [getattr(SAILR_DECOMPILERS, a) for a in dir(SAILR_DECOMPILERS) if not a.startswith("_")]
# TODO: this can be bad if people run it not in our pipeline
JOERNLESS_SERVER_METRICS = [SAILR_METRICS.CFGED, SAILR_METRICS.GED_UPPERBOUND, SAILR_METRICS.GED_MAX,
                            SAILR_METRICS.GED_EXACT, SAILR_METRICS.BLOCK_COUNT, SAILR_METRICS.DEC_TIME,
                            SAILR_METRICS.CODE_COMPLEXITY]




import os
import time
from multiprocessing.pool import Pool
from pathlib import Path
from subprocess import run
import shutil
import logging
import re
from typing import Dict
import traceback

import toml
from tqdm import tqdm
from pyjoern import JoernClient, JoernServer

from .sailr_target import SAILRTarget
from ..utils import timeout
from . import (
    bcolors, WorkDirContext, SAILR_COMPILATION_RESULTS_DIR, SAILR_DECOMPILATION_RESULTS_DIR,
    SAILR_MEASURE_RESULTS_DIR
)
from .. import SAILR_METRICS

l = logging.getLogger(__name__)

#
# compile flags
#

DEFAULT_OPTIMIZATION_LEVELS = ["O2"]
OPTIMIZATION_LEVELS = ["O0", "O1", "O2", "O3"]

# old debug flags
DUMP_GED_SOURCE = "-fno-inline -fdump-tree-original"
DUMP_FLAGS = "-fdump-tree-all-details -fdump-rtl-all"

# base flag, always start with this
BASE_FLAGS = f"-g -fno-inline -fno-builtin -save-temps=obj"
# Jump Threading
NO_JUMP_THREAD = "-fno-thread-jumps -fno-tree-vrp -fno-branch-target-load-optimize -fno-branch-target-load-optimize2 " \
                 "-fno-tree-dominator-opts -fdisable-tree-ethread -fdisable-tree-thread1 -fdisable-tree-thread2 " \
                 "-fdisable-tree-thread3 -fdisable-tree-thread4"
# Common Subexpression Elimination (run the 3 below together as one)
NO_CSE = "-fno-cse-skip-blocks -fno-cse-follow-jumps -fno-rerun-cse-after-loop -fdisable-rtl-cse1 -fdisable-rtl-cse2"
NO_GCSE = "-fno-gcse -fno-gcse-lm -fdisable-rtl-gcse2"
NO_PRE = "-fno-tree-pre"  # partial redundancy elimination (a type of CSE)

# Switch Conversion
NO_SWITCH_CONV = "-fno-tree-switch-conversion -fno-tree-tail-merge"

# Cross Jumping
NO_CROSS_JUMPING = "-fno-crossjumping -fdisable-rtl-jump -fdisable-rtl-jump2"

# Software-Thread Cache Basic Block Reordering
NO_STC_BBRO = "-freorder-blocks-algorithm=simple -fno-reorder-blocks"

# Loop Header & Invariant Optimizations
NO_LOOP_HEADER = "-fno-tree-ch"
NO_LOOP_INV = "-fno-tree-loop-optimize -fdisable-rtl-loop2_invariant"

# Forward Store Motion
NO_TREE_SINK = "-fno-tree-sink"

# No Builtins (a type of inlining)
NO_BUILTIN = "-fno-builtin"

# DISABLE EVERYTHING!
NO_EVERYTHING = f"{BASE_FLAGS} {DUMP_FLAGS} {NO_JUMP_THREAD} {NO_CSE} {NO_GCSE} {NO_PRE} {NO_SWITCH_CONV} {NO_CROSS_JUMPING} {NO_STC_BBRO} {NO_LOOP_HEADER} {NO_LOOP_INV} {NO_TREE_SINK}"

# Then do:
# - No Switch Functions
# - No Gotos from Source
# - No IDA failures

PKG_COPY_BLACKLIST = {"kmod",} 

def sailr_target_src_exists(sailr_target: SAILRTarget, opt_level="O2") -> bool:
    results_dir = Path(f"./results/{opt_level}/").absolute()
    package_dir = results_dir.joinpath(sailr_target.package_dir).absolute()
    return package_dir.exists() 


def compile_and_copy_clean_files(sailr_target: SAILRTarget, cflags, cores, opt_levels=DEFAULT_OPTIMIZATION_LEVELS,  cc='gcc', cxx='g++', clean=False):
    # if any targets are not existent, check to see if we can copy files over
    if sailr_target.download and not all(sailr_target_src_exists(sailr_target, opt_level=lvl) for lvl in opt_levels):
        l.info("Detecting some optimization levels need files for compiling. Checking other levels...")

        if sailr_target.package_name not in PKG_COPY_BLACKLIST:
            # find an already downloaded target, do caching
            for cached_opt_level in opt_levels:
                if sailr_target_src_exists(sailr_target, opt_level=cached_opt_level):
                    break
            else:
                # if none exists then we download one
                cached_opt_level = opt_levels[0]
                setup_sailr_compile_target(sailr_target, cached_opt_level)

            for opt_level in opt_levels:
                if opt_level == cached_opt_level:
                    continue
                
                # copy files over for each target
                l.info(f"Copying files from {cached_opt_level} to {opt_level}...")
                copy_compiled_files_from_opt(sailr_target, cached_opt_level, opt_level)
        else:
            # blacklist packages require all files to be setup :( 
            for opt_level in opt_levels:
                setup_sailr_compile_target(sailr_target, opt_level)

    # compile the actual files now
    for opt_level in opt_levels:
        compile_target(
            sailr_target, cflags, cores, opt_level, cc=cc, cxx=cxx, 
            setup_target=not sailr_target_src_exists(sailr_target, opt_level=opt_level)
        )
        if sailr_target.skip_file_copy:
            continue

        compiled_res_dir = copy_compiled_files(sailr_target, opt_level)
        if compiled_res_dir is None:
            continue


def clean_and_update_source_from_i(compiled_res_dir, sailr_target: SAILRTarget, save_linemaps=True):
    """
    Only works with *.i files are in the directory given, otherwise no work is done.

    :param compiled_res_dir:
    :param save_linemaps:
    :return:
    """
    l.info("Cleaning and mapping source lines now. This may take a minute...")
    cleaned_files_ln_map = clean_target_i_files(compiled_res_dir, sailr_target)
    if save_linemaps:
        if not cleaned_files_ln_map:
            l.info("There are no files to clean...")
            return
        correct_and_save_line_maps(cleaned_files_ln_map)


def _collect_function_lines(
    src_file, i_file, known_linemap, joern_port=9000,
    metrics=[SAILR_METRICS.GOTO_COUNT, SAILR_METRICS.BOOLEAN_COUNT, SAILR_METRICS.FUNC_CALLS],
    do_tiemout=True,
):
    from ..analysis.measure import measure
    metric_amount = len(metrics) if metrics else 0
    cachable_results = {}
    basename = i_file.with_suffix("").with_suffix("").name

    def _eval_with_server(joern_port, src_file, i_file):
        with JoernServer(port=joern_port):
            src_client = JoernClient(src_file, port=joern_port)
            src_ln_map = src_client.function_line_numbers()
            src_functions = src_client.functions

        if not src_functions:
            raise ValueError(f"Could not find any functions in {src_file}")

        with JoernServer(port=joern_port + 1):
            client = JoernClient(i_file, port=joern_port + 1, bin_name=basename)
            i_ln_map = client.function_line_numbers()
            extra_metrics = {SAILR_METRICS.DEC_TIME: 0.0}
            cache_res = measure(i_file, basename, metrics, src_functions, client=client, **extra_metrics)

        return cache_res, src_functions, i_ln_map, src_ln_map

    try:
        if do_tiemout:
            with timeout(seconds=100 + 20 * metric_amount):
                cachable_results[basename], src_functions, i_ln_map, src_ln_map = _eval_with_server(joern_port, src_file, i_file)
        else:
            cachable_results[basename], src_functions, i_ln_map, src_ln_map = _eval_with_server(joern_port, src_file, i_file)
    except (TimeoutError, Exception) as e:
        l.info(f"JOERN timeout for {src_file}! with {e}")
        traceback.print_exception(e)
        return None, None, None, None

    final_linemap = {}
    for function, (src_start, src_end) in src_ln_map.items():
        func_linemap = {}
        i_data = i_ln_map.get(function, None)
        if i_data is None:
            continue

        i_start, i_end = i_data
        src_pos = src_start
        for ln in range(i_start, i_end+1):
            # skip if known
            if ln in known_linemap:
                func_linemap[ln] = known_linemap[ln]
                continue

            func_linemap[ln] = src_pos
            src_pos += 1

        final_linemap[function] = func_linemap

    str_linemaps = {}
    for func, mapping in final_linemap.items():
        str_linemaps[func] = {str(k): v for k,v in mapping.items()}

    name = i_file.with_suffix("").with_suffix("").name
    out_file = i_file.parent.joinpath(f"source_{name}.linemaps")
    with open(out_file, "w") as fp:
        toml.dump(str_linemaps, fp)

    with open(i_file.parent.joinpath(f"source_{name}.toml"), "w") as fp:
        toml.dump(cachable_results[basename], fp)

    return i_file, i_ln_map, known_linemap, src_ln_map


def correct_and_save_line_maps(cleaned_files_lines_map: Dict, cores=15, joern_port_start=9000):
    # 1. get line start and end for every function, by name, in each file (*.c and *.i)
    pairs = []
    for cleaned_file, line_map in cleaned_files_lines_map.items():
        src_path = cleaned_file.with_suffix("").with_suffix(".c")
        alt_src_path = cleaned_file.with_suffix("").with_suffix(".orig.c")
        is_alt = False
        if not src_path.exists():
            is_alt = True
            if not alt_src_path.exists():
                continue
            src_path = alt_src_path

        if not is_alt:
            # move .c -> .orig.c (for parsing)
            shutil.move(src_path, src_path.with_suffix(".orig.c"))

        # copy .i.c -> .c (for later pipeline)
        shutil.copy(cleaned_file, src_path)

        if not is_alt:
            src_path = src_path.with_suffix(".orig.c")
        pairs.append((src_path, cleaned_file, line_map))

    linemaps = {}
    with Pool(processes=cores) as pool, tqdm(total=len(pairs)) as pbar:
        work = []
        for i, pair in enumerate(pairs):
            work.append(pool.apply_async(
                _collect_function_lines,
                args=(pair[0], pair[1], pair[2]),
                kwds={"joern_port": joern_port_start+2*i},
                callback=lambda _: pbar.update(1)
            ))

        pool.close()
        for w in work:
            try:
                out = w.get(timeout=200)
                if not out:
                    pbar.update(1)
                    continue

                if len(out) < 4:
                    pbar.update(1)
                    continue

                f, i_ln_map, known_linemap, src_ln_map = out
                if f is None:
                    pbar.update(1)
                    continue

                linemaps[f] = (i_ln_map, known_linemap, src_ln_map)
            except Exception as e:
                print(e)
                pbar.update(1)
                continue

    # 2. We now have all the lines. Starting at each function traverse start to finish, ignoring places we already
    # know from the known_linemap
    linemap_by_file = {}
    for filepath, (i_ln_map, known_linemap, src_ln_map) in linemaps.items():
        final_linemap = {}
        for function, (src_start, src_end) in src_ln_map.items():
            func_linemap = {}
            i_data = i_ln_map.get(function, None)
            if i_data is None:
                continue

            i_start, i_end = i_data
            src_pos = src_start
            for ln in range(i_start, i_end+1):
                # skip if known
                if ln in known_linemap:
                    func_linemap[ln] = known_linemap[ln]
                    continue

                func_linemap[ln] = src_pos
                src_pos += 1

            final_linemap[function] = func_linemap

        linemap_by_file[filepath] = final_linemap

    if not linemap_by_file:
        l.critical("Nonexistent linemaps for all files!")

    # 3. save all the linemaps
    for i_c_filepath, linemap in linemap_by_file.items():
        str_linemaps = {}
        for func, mapping in linemap.items():
            str_linemaps[func] = {str(k): v for k,v in mapping.items()}

        name = i_c_filepath.with_suffix("").with_suffix("").name
        out_file = i_c_filepath.parent.joinpath(f"source_{name}.linemaps")
        with open(out_file, "w") as fp:
            toml.dump(str_linemaps, fp)

    return linemap_by_file


def clean_target_i_files(compiled_results_dir: Path, sailr_target: SAILRTarget):
    files = list(compiled_results_dir.rglob("*.i"))
    cleaned_file_ln_map = {}
    for file in files:
        new_file, ln_map = clean_i_file(file.absolute(), sailr_target)
        cleaned_file_ln_map[new_file] = ln_map

    return cleaned_file_ln_map


def clean_i_file(filepath: Path, sailr_target: SAILRTarget):
    filepath = Path(filepath)
    if sailr_target and sailr_target.windows_project:
        line_comment_str = b'#line ([0-9]+) \"(.+)\".*'
    else:
        line_comment_str = b'# ([0-9]+) \"(.+)\".*'

    with open(filepath, "rb") as fp:
        data = fp.read()

    data = re.sub(b"__extension__ extern [^;]*;\n", b"", data, flags=re.DOTALL)
    data = re.sub(b"extern [^;]*;\n", b"", data, flags=re.DOTALL)
    data = re.sub(b"__extension__\n", b"", data, flags=re.DOTALL)
    lines = data.split(b"\n")
    clean_lines = list()
    for i in range(1, len(lines)):
        prev_line = lines[i-1]
        curr_line = lines[i]

        # skip curr line with debug
        has_debug = re.match(b"#.*", curr_line)
        if has_debug is not None and curr_line.startswith(b"#"):
            continue

        prev_has_debug = re.match(line_comment_str, prev_line)
        line_num = None
        if prev_has_debug is not None and prev_line.startswith(b"#") and Path(prev_has_debug[2].decode()).with_suffix("").name == filepath.with_suffix("").name:
            line_num = int(prev_has_debug[1])

        clean_lines.append((curr_line, line_num))

    # reconstruct the data
    data = b""
    corrected_linenumbers = {}
    for i, (line, linenum) in enumerate(clean_lines):
        data += line + b"\n"
        if linenum is not None:
            corrected_linenumbers[i] = linenum

    #
    # removes special comments found in GCC preprocessing:
    # https://gcc.gnu.org/onlinedocs/cpp/Preprocessor-Output.html
    #
    #data = re.sub(b"#.*\n", b"", data)
    #data = re.sub(b"\\[\\[.*\\]\\]", b"", data)

    with open(filepath, "wb") as fp:
        fp.write(data)

    #
    # make the output look decent with clang-format
    #
    #run(f"clang-format -i {filepath}".split())

    new_file = filepath.with_suffix(".i.c")
    shutil.move(filepath, new_file)

    return new_file, corrected_linenumbers


def copy_compiled_files(sailr_target: SAILRTarget, opt_level):
    target_dir = Path(f"./results/{opt_level}/").joinpath(sailr_target.package_dir).absolute()
    if not target_dir.exists():
        l.warning(f"Was unable to find the package directory for {sailr_target.package_name}")
        return None

    target_src_dir = target_dir.joinpath(sailr_target.source_dir).absolute()
    if not target_src_dir.exists():
        l.warning(f"was unable to find the source directory for package {sailr_target.package_name} for {sailr_target.source_dir}")
        return None

    # make compilation output dir if it does not exist
    sailr_comp_dir = target_dir.joinpath(SAILR_COMPILATION_RESULTS_DIR).absolute()
    if sailr_comp_dir.exists():
        shutil.rmtree(sailr_comp_dir, ignore_errors=True)
    os.mkdir(sailr_comp_dir)

    # copy .i, .c, and .o files
    file_ends = (".i", ".c", ".o")
    for file_end in file_ends:
        files = list(target_src_dir.rglob(f"*{file_end}"))
        for file in files:
            new_loc = sailr_comp_dir.joinpath(file.name).absolute()
            if sailr_target.rename_dash_files:
                if "-" in new_loc.name:
                    new_loc = new_loc.with_name("-".join(new_loc.name.split("-")[1:]))

            if new_loc.exists():
                l.warning(f"Overwriting a file {file.name} that already exists")

            shutil.copy(file.absolute(), new_loc)

    # remove files without all 3!
    bad_files = set()
    for file_end in file_ends:
        other_ends = [end for end in file_ends if end != file_end]
        for file in list(sailr_comp_dir.rglob(f"*{file_end}")):
            if not file.with_suffix(other_ends[0]).exists() or not file.with_suffix(other_ends[1]).exists():
                bad_files.add(file.absolute())

    for file in bad_files:
        l.warning(f"Removing {file} because it is missing other files")
        os.unlink(file)

    good_files = sorted(list(sailr_comp_dir.rglob(f"*.o")))
    if sailr_target.subset_compilation:
        bad_files = good_files[sailr_target.subset_compilation:]
        for file in bad_files:
            os.unlink(file)
            os.unlink(file.with_suffix(".i"))
            os.unlink(file.with_suffix(".c"))

    return sailr_comp_dir


def copy_compiled_files_from_opt(sailr_target: SAILRTarget, src_opt_level, dst_opt_level):
    src_dir = Path(f"./results/{src_opt_level}/").joinpath(sailr_target.package_dir).absolute()
    dst_dir = Path(f"./results/{dst_opt_level}/").joinpath(sailr_target.package_dir).absolute()
    l.info(f"Using cached download to copy files from {src_dir} to {dst_dir}")
    shutil.copytree(src_dir, dst_dir, ignore_dangling_symlinks=True)


def setup_sailr_compile_target(sailr_target: SAILRTarget, opt_level):
    l.info(f"Downloading files for {sailr_target.package_name}/{opt_level}...")
    if opt_level not in OPTIMIZATION_LEVELS:
        l.critical(f"Invalid optimization level {opt_level}, quiting...")
        return

    results_dir = Path(f"./results/{opt_level}/").absolute()
    # download and unpack the target if download is on for the target
    if sailr_target.download:
        with WorkDirContext(results_dir):
            package_dir_top = ""
            if sailr_target.package_dir:
                package_dir_top = Path(sailr_target.package_dir)
                # only use the parent if you see subdirs in the package folder
                if "/" in sailr_target.package_dir:
                    package_dir_top = package_dir_top.parent

            filename = sailr_target.source_remote.split("/")[-1]
            if sailr_target.remote_type.lower() == "git":
                run(f"git clone {sailr_target.source_remote} {package_dir_top}".split())
            elif sailr_target.remote_type.lower() == "zip":
                run(f"wget -nc {sailr_target.source_remote}".split())
                run(f"unzip -o {filename}".split())
                run(f"rm {filename}".split())
            elif sailr_target.remote_type.lower() == "tar":
                run(f"wget -nc {sailr_target.source_remote}".split())
                run(f"tar xvf {filename}".split())
                run(f"rm {filename}".split())
            elif sailr_target.remote_type.lower() == "local":
                local_folder = Path(sailr_target.source_remote).absolute()
                shutil.copytree(local_folder, sailr_target.package_dir, ignore_dangling_symlinks=True)
            else:
                print(f"unsupported or missing remote file type: {sailr_target.remote_type}, {sailr_target.source_remote}")

    # now we are inside the package
    with WorkDirContext(results_dir.joinpath(sailr_target.package_dir)):
        # if download was run in this execution, we must attempt a checkout
        # if the remote type is git; we may also need to do post-download things
        if sailr_target.download:
            # checkout to the correct git version
            if sailr_target.version and sailr_target.remote_type.lower() == "git":
                run(f"git fetch --tags".split())
                run(f"git checkout tags/{sailr_target.version} -b sailr_eval".split())

            # run post-download commands
            if sailr_target.post_download_cmds:
                for cmd in sailr_target.post_download_cmds:
                    run(cmd.split())

            # apply patch if it has one
            if sailr_target.apply_patch:
                patch_path = results_dir.joinpath("patches").joinpath(sailr_target.apply_patch).absolute()
                if patch_path.exists():
                    run(f"git apply {patch_path}".split())


def compile_target(sailr_target: SAILRTarget, cflags, cores, opt_level, cc='gcc', cxx='g++', setup_target=True):
    if sailr_target.skip_compilation:
        l.warning("This project is marked to skip automated compilation. Skipping...")
        return

    if sailr_target.windows_project:
        l.warning("Windows projects must be compiled manually. Assuming files are in this dir and copying them...")
        return

    if setup_target:
        setup_sailr_compile_target(sailr_target, opt_level)

    results_dir = Path(f"./results/{opt_level}/").absolute()
    with WorkDirContext(results_dir.joinpath(sailr_target.package_dir)):
        # start setting up the environment for compiling
        # make a copy of the current env so we can edit it later
        d = dict(os.environ)
        d['CC'] = sailr_target.c_compiler or cc
        d['CXX'] = sailr_target.cpp_compiler or cxx
        d['CFLAGS'] = f"-{opt_level} " + cflags
        d['CPPFLAGS'] = d['CFLAGS']

        print(f"{bcolors.WARNING}{bcolors.BOLD} Configuring with flags: {bcolors.ENDC} {d['CFLAGS']} for {d['CC']}")
        time.sleep(1)

        if sailr_target.pre_make_cmds:
            for cmd in sailr_target.pre_make_cmds:
                print(f"running {cmd}")
                ret = run(cmd.split(), env=d)
                # verify things were configured!
                if ret.returncode != 0:
                    print("Failed to configure... please check the log file file config.log")

        print(f"{bcolors.WARNING}{bcolors.BOLD} Configuration successful, compiling now... {bcolors.ENDC} {d['CFLAGS']}")
        time.sleep(1)

        run(f"make clean".split(), env=d)
        if sailr_target.override_make_vars:
            base_make_cmd = sailr_target.make_cmd.split()
            new_make_cmd = base_make_cmd[:1] + [f'CFLAGS={cflags}'] + base_make_cmd[1:]
            new_make_cmd += [f'-j{cores}'] if cores != 1 else []
            run(new_make_cmd, env=d)
        else:
            run(f"{sailr_target.make_cmd} {'-j'+str(cores) if cores != 1 else ''}".split(), env=d)

        if sailr_target.post_make_cmds:
            run(sailr_target.post_make_cmds.split())


def strip_file(filepath: Path):
    run(f"strip --strip-debug {filepath}".split())



import logging
from pathlib import Path
from typing import Dict

from pyjoern.mapping import correct_decompiler_mappings, read_line_maps
from cfgutils.similarity import ged_max, ged_upperbound, ged_exact
from cfgutils.similarity import cfg_edit_distance as _cfg_edit_distance
import networkx as nx

l = logging.getLogger(__name__)
MAX_EXACT_NODES = 12

#
# Graph Edit Distance Variants
#


def ged_exact_score(
    func_name, client, source_cfgs: Dict[str, nx.DiGraph] = None, dec_cfgs: Dict[str, nx.DiGraph] = None,
    decompiler=None, binary_path=None, **kwargs
):
    """
    Computes the exact GED for a given function.
    """
    if decompiler == "source":
        return float(0)
    source_cfg, dec_cfg = _verify_has_valid_graphs(func_name, client, source_cfgs, dec_cfgs, decompiler, binary_path)
    if source_cfg is None or dec_cfg is None:
        return None

    return ged_exact(source_cfg, dec_cfg)


def ged_upperbound_score(
    func_name, client, source_cfgs: Dict[str, nx.DiGraph] = None, dec_cfgs: Dict[str, nx.DiGraph] = None,
    decompiler=None, binary_path=None, **kwargs
):
    """
    Collects the upperbound fastest approximation of GED for a given function. This is a highly innacurate metric.
    """
    if decompiler == "source":
        return float(0)
    source_cfg, dec_cfg = _verify_has_valid_graphs(func_name, client, source_cfgs, dec_cfgs, decompiler, binary_path)
    if source_cfg is None or dec_cfg is None:
        return None

    return ged_upperbound(source_cfg, dec_cfg)


def ged_max_score(
    func_name, client, source_cfgs: Dict[str, nx.DiGraph] = None, dec_cfgs: Dict[str, nx.DiGraph] = None,
    binary_path=None, decompiler=None, **kwargs
):
    """
    Computes the MAXIMUM possible score that GED could return for a given function.
    Notice, this score is just a summation of all edges and nodes from both graphs.
    """
    source_cfg, dec_cfg = _verify_has_valid_graphs(func_name, client, source_cfgs, dec_cfgs, decompiler, binary_path)
    if source_cfg is None or dec_cfg is None:
        return None

    return ged_max(source_cfg, dec_cfg)


#
# CFG Edit Distance
#


def cfg_edit_distance(
    func_name, client, source_cfgs: Dict[str, nx.DiGraph] = None, dec_cfgs: Dict[str, nx.DiGraph] = None,
    binary_path=None, decompiler=None, **kwargs
):
    if decompiler == "source":
        return float(0)
    source_cfg, dec_cfg = _verify_has_valid_graphs(func_name, client, source_cfgs, dec_cfgs, decompiler, binary_path)
    if source_cfg is None or dec_cfg is None:
        return None

    return compute_cfg_edit_distance(dec_cfg, source_cfg, func_name, binary_path, decompiler)


def compute_cfg_edit_distance(dec_cfg, src_cfg, function, binary_path: Path, decompiler: str):
    binary_path = Path(binary_path)
    bin_dir = binary_path.parent
    bin_name = binary_path.with_suffix("").name

    src_addr_to_line_map = read_line_maps(binary_path.with_suffix(".dwarf.linemaps"), key_is_func=False, value_is_set_type=False)
    if not src_addr_to_line_map:
        return None

    dec_line_to_addr_map = read_line_maps(bin_dir.joinpath(f"{decompiler}_{bin_name}.linemaps"))
    if not dec_line_to_addr_map:
        return None
    func_line_to_addr_map = dec_line_to_addr_map.get(function, None)
    if func_line_to_addr_map is None:
        return None

    dec_line_to_addr_map = correct_decompiler_mappings(dec_cfg, func_line_to_addr_map)
    return _cfg_edit_distance(dec_cfg, src_cfg, dec_line_to_addr_map, src_addr_to_line_map)


def has_cfged_required_src_files(source_path: Path, binary_path: Path):
    dwarf_linemaps = binary_path.with_suffix(".dwarf.linemaps").absolute()
    return dwarf_linemaps.exists() and has_cfged_required_files(source_path)


def has_cfged_required_files(source_path: Path):
    source_linemaps = source_path.with_suffix(".linemaps").absolute()
    return source_linemaps.exists()

#
# Others
#


def _verify_has_valid_graphs(
    func_name, client, source_cfgs: Dict[str, nx.DiGraph], dec_cfgs: Dict[str, nx.DiGraph], decompiler, binary_path
):
    if client is not None:
        if func_name not in client.functions:
            return None, None

        if not source_cfgs or not dec_cfgs:
            return None, None

    source_cfg = source_cfgs.get(func_name, None)
    dec_cfg = dec_cfgs.get(func_name, None)
    if source_cfg is None or dec_cfg is None:
        l.debug(f"CFGs are none for {decompiler} on {func_name} in {binary_path}")
        return None, None
    elif len(source_cfg.nodes) == 0:
        l.debug(f"Empty CFG on source for {func_name} in {binary_path}")
        return None
    elif len(dec_cfg.nodes) == 0:
        l.debug(f"Empty CFG on {decompiler} for {func_name} in {binary_path}")
        return None, None

    return source_cfg, dec_cfg


def graph_size(
    func_name, client, source_cfgs: Dict[str, nx.DiGraph] = None, dec_cfgs: Dict[str, nx.DiGraph] = None,
    binary_path=None, decompiler=None, **kwargs
):
    dec_cfg = dec_cfgs.get(func_name, None)
    if dec_cfg is None:
        return None

    return len(dec_cfg.nodes) + len(dec_cfg.edges)


def block_count(
    func_name, client, source_cfgs: Dict[str, nx.DiGraph] = None, dec_cfgs: Dict[str, nx.DiGraph] = None,
    binary_path=None, decompiler=None, **kwargs
):
    dec_cfg = dec_cfgs.get(func_name, None)
    if dec_cfg is None:
        l.debug(f"CFGs are none for {decompiler} on {func_name} in {binary_path}")
        return None

    return len(dec_cfg.nodes)



import networkx as nx

def get_graph_root_nodes(graph: nx.DiGraph, **kwargs):
    initial_roots = [n for n in graph.nodes if graph.in_degree(n) == 0]
    roots = []
    for ir in initial_roots:
        label = graph._node[ir]['label']
        if label.startswith(f"(METHOD,{graph.name}") and "SUB" in label:
            roots.append(ir)

    return roots


from pathlib import Path
import textwrap
from typing import Dict, List
import matplotlib as mpl
mpl.use('Agg')
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
#import seaborn as sns

STAT_TYPES = ("sum", "mean", "median")



#
# Graphs
#

def plot_histogram(data1, data2, save_path="./histo_chart.png"):
    import matplotlib.pyplot as plt
    import numpy as np
    
    # Sample Data
    IDA = data1
    SAILR = data2
    # Calculate the unique elements and their counts for both lists
    values_IDA, counts_IDA = np.unique(IDA, return_counts=True)
    values_SAILR, counts_SAILR = np.unique(SAILR, return_counts=True)
    
    # Create dictionaries for both datasets
    dict_IDA = dict(zip(values_IDA, counts_IDA))
    dict_SAILR = dict(zip(values_SAILR, counts_SAILR))
    
    # Ensure every key in one dictionary exists in the other (even if it's zero)
    for key in dict_IDA.keys():
        dict_SAILR[key] = dict_SAILR.get(key, 0)
    
    for key in dict_SAILR.keys():
        dict_IDA[key] = dict_IDA.get(key, 0)
    
    # Sort the dictionaries by their keys for plotting
    dict_IDA = dict(sorted(dict_IDA.items()))
    dict_SAILR = dict(sorted(dict_SAILR.items()))
    
    # Calculate the difference in counts
    diff = [dict_SAILR[key] - dict_IDA[key] for key in dict_IDA.keys()]
    
    # Plot
    plt.bar(dict_IDA.keys(), diff, color=['red' if d < 0 else 'blue' for d in diff])
    plt.axhline(0, color='black',linewidth=0.8)
    plt.xlabel('CFGED Score')
    plt.ylabel('SAILR count - IDA count')
    plt.ylim(-60, 60)  # Adjusting the upper limit to +

    plt.title('Difference in frequencies between SAILR and IDA')

    plt.savefig(save_path)


def plot_diff_barchar(data_by_row: Dict[str, List], metric: str, decompilers: List, save_path="./diff_bar_chart.png"):
    assert len(decompilers) == 2
    metric_data = data_by_row[metric]

    data = {
        'Function': range(len(metric_data[decompilers[0]])),
        decompilers[0]: metric_data[decompilers[0]],
        decompilers[1]: metric_data[decompilers[1]]
    }
    df = pd.DataFrame(data)
    df['Diff'] = df[decompilers[0]] - df[decompilers[1]]
    df['Diff'] = df['Diff'].where(df['Diff'] > 0)

    df = df.sort_values('Diff', ascending=False)
    df['Function'] = range(df['Diff'].size)
    plt.bar(df['Function'], df['Diff'], label=f'{decompilers[0]} - {decompilers[1]}')
    plt.xlabel('Function')
    plt.ylabel(metric)
    plt.savefig(save_path)


def plot_barchart(data_by_row: Dict[str, List], save_path="./bar_chart.png"):
    data = {
        'Function': range(len(data_by_row["angr_sailr"])),
        'angr_phoenix': data_by_row["angr_phoenix"],
        'angr_sailr': data_by_row["angr_sailr"]
    }
    df = pd.DataFrame(data)

    # Step 3: Sort your data
    # Let's say we want to sort by the total of 'Value1' and 'Value2' in descending order
    df['Diff'] = df['angr_sailr'] - df['angr_phoenix']
    df = df.sort_values('Diff', ascending=False)
    df['Function'] = range(len(df['angr_phoenix']))
    plt.bar(df['Function'], df['Diff'], label='angr_sailr - angr_phoenix')

    # Set the x and y-axis labels
    plt.xlabel('Function')
    plt.ylabel('CFGED')

    # Set a title
    plt.title('CFGED Scores')

    plt.savefig(save_path)


def plot_box_plot(data_by_row: Dict[str, List], save_path="./boxplot.png"):
    plt.figure(figsize=(10, 6))  # Set the figure size
    plt.boxplot(data_by_row.values(), vert=False, widths=0.7)  # Plot the data

    # Set the y-axis labels
    plt.yticks([1, 2], data_by_row.keys())

    # Set the x-axis limits to give some margins
    plt.xlim([-50, 1450])  # You can adjust these numbers based on your data

    # Set the x-axis label
    plt.xlabel('Values')

    # Set a title
    plt.title('Boxplot of Data')

    plt.savefig(save_path)


#
# Markdown Table
#

def tablify_data_for_metrics(summary):
    first_dec = list(summary.keys())[0]
    all_metrics = list(summary[first_dec].keys())
    good_metrics = set()
    if "metadata" in summary:
        del summary["metadata"]
    for metric in all_metrics:
        raw_metric_name = "_".join(metric.split("_")[:-1])
        is_bad = False
        for stat_type in STAT_TYPES:
            if f"{raw_metric_name}_{stat_type}" not in all_metrics:
                is_bad = True
                break

        if not is_bad:
            good_metrics.add(raw_metric_name)

    row_by_metric = list()
    for metric in good_metrics:

        row = {"Metric": metric}
        for dec in summary.keys():
            data = []
            for stat_type in STAT_TYPES:
                data.append(summary[dec][f"{metric}_{stat_type}"])

            #row[dec] = str(tuple(data))
            row[dec] = "/".join([str(d) for d in data])

        row_by_metric.append(row)

    return row_by_metric


def tablify_data(summary, show_stats=False):
    table_summary = list()
    metadata_table = {}
    for dec, data in summary.items():

        row = {"Decompiler": dec}
        for metric_name, metric_data in data.items():
            row[metric_name] = metric_data

        if dec == "metadata":
            del row["Decompiler"]
            metadata_table = row
            continue

        table_summary.append(row)

    if show_stats:
        table_summary = tablify_data_for_metrics(summary)

    return table_summary, metadata_table


def save_data_as_markdown_table(summary, save_path="./summary.md", show_stats=False):
    def format_table(myDict, colList=None):
        """ Pretty print a list of dictionaries (myDict) as a dynamically sized table.
        If column names (colList) aren't specified, they will show in random order.
        Author: Thierry Husson - Use it as you want but don't blame me.
        """
        if not colList: colList = list(myDict[0].keys() if myDict else [])
        myList = [colList]  # 1st row = header
        for item in myDict: myList.append([str(item[col] if item[col] is not None else '') for col in colList])
        colSize = [max(map(len, col)) for col in zip(*myList)]
        formatStr = ' | '.join(["{{:<{}}}".format(i) for i in colSize])
        myList.insert(1, ['-' * i for i in colSize])  # Seperating line
        for item in myList: yield formatStr.format(*item)

    markdown_summary = ""
    for pkg_name, pkg_data in summary.items():
        table_summary, metadata_table = tablify_data(pkg_data, show_stats=show_stats)
        markdown_summary += f"### {pkg_name}\n" + f"\n".join(list(format_table(table_summary))) + "\n\n"

    markdown_metadata = "\n".join(list(format_table([metadata_table])))
    markdown_doc = textwrap.dedent(f"""\n\n# Evaluation Data
## Stats
{f"Layout: {STAT_TYPES}" if show_stats else ""}
{markdown_summary}
## Metadata

{markdown_metadata}""")

    with open(save_path, "w") as fp:
        fp.write(markdown_doc)

    return markdown_doc



"""
Counts the amount of nesting for each if-statement
"""

def count_nesting_ifs(func_name, client, **kwargs):
    if client is None:
        return None

    return client.count_if_levels(func_name)


from typing import Dict, List, Tuple

from .. import SAILR_DECOMPILERS, SAILR_METRICS


def run_post_metrics(post_metrics: List[str], metric_list_by_decompiler: Dict[str, Dict[str, List[int]]]) -> Dict[str, Dict[str, List[int]]]:
    from . import POST_METRICS
    """
    Runs the post metrics on the metric_list_by_decompiler.
    """
    final_output = metric_list_by_decompiler.copy()
    for metric in post_metrics:
        if metric not in POST_METRICS:
            raise ValueError(f"Unknown post metric: {metric}")

        round_output = POST_METRICS[metric](final_output)
        for dec in round_output:
            final_output[dec][metric] = round_output[dec]

    return final_output


def goto_func_cfged(metric_list_by_decompiler: Dict[str, Dict[str, List[int]]]) -> Dict[str, Dict[str, List[int]]]:
    output, valid = _validate_metric_data([SAILR_METRICS.CFGED, SAILR_METRICS.GOTO_COUNT], metric_list_by_decompiler)
    if not valid or SAILR_DECOMPILERS.ANGR_PHOENIX not in metric_list_by_decompiler:
        return output

    # find functions with goto >= 1 in Phoenix
    goto_idxs = []
    for i, goto_count in enumerate(metric_list_by_decompiler[SAILR_DECOMPILERS.ANGR_PHOENIX][SAILR_METRICS.GOTO_COUNT]):
        if goto_count >= 1:
            goto_idxs.append(i)

    # get CFGED for those functions
    for dec, mdata in metric_list_by_decompiler.items():
        for i, cfged in enumerate(mdata[SAILR_METRICS.CFGED]):
            if i in goto_idxs:
                output[dec].append(cfged)

    return output


def norm_cfged(metric_list_by_decompiler: Dict[str, Dict[str, List[int]]]) -> Dict[str, Dict[str, List[int]]]:
    """
    Computes: CFGED / GRAPH_SIZE
    Requires: CFGED, GRAPH_SIZE
    """
    output, valid = _validate_metric_data([SAILR_METRICS.CFGED, SAILR_METRICS.GRAPH_SIZE], metric_list_by_decompiler)
    if not valid:
        return output

    for dec, mdata in metric_list_by_decompiler.items():
        for i, cfged in enumerate(mdata[SAILR_METRICS.CFGED]):
            output[dec].append(cfged / metric_list_by_decompiler[SAILR_DECOMPILERS.SOURCE_CODE][SAILR_METRICS.GRAPH_SIZE][i])

    return output


def zero_cfged(metric_list_by_decompiler: Dict[str, List[int]]) -> Dict[str, List[int]]:
    output, valid = _validate_metric_data([SAILR_METRICS.CFGED], metric_list_by_decompiler)
    if not valid:
        return output

    for dec, mdata in metric_list_by_decompiler.items():
        for cfged in mdata[SAILR_METRICS.CFGED]:
            output[dec].append(1 if cfged == 0 else 0)

    return output


def funcs_w_goto(metric_list_by_decompiler: Dict[str, List[int]]) -> Dict[str, List[int]]:
    output, valid = _validate_metric_data([SAILR_METRICS.GOTO_COUNT], metric_list_by_decompiler)
    if not valid:
        return output

    for dec, mdata in metric_list_by_decompiler.items():
        for gotos in mdata[SAILR_METRICS.GOTO_COUNT]:
            output[dec].append(1 if gotos >= 1 else 0)

    return output

#
# Utils
#

def _validate_metric_data(requires: List[str], metric_list_by_decompiler) -> Tuple[Dict, bool]:
    """
    Validates that the required metrics are in the metric_list_by_decompiler.
    """
    output = {dec: list() for dec in metric_list_by_decompiler}
    for dec, mdata in metric_list_by_decompiler.items():
        if any(m not in mdata for m in requires):
            return output, False

    return output, True



import os
import sys
import shutil
from typing import List, Dict, Tuple

from .angr_dec import angr_decompile, angr_disassemble, get_all_funcs, angr_phoenix_decompile, angr_dream_decompile, \
    angr_improved_phoenix_decompile, angr_comb_decompile
from .ida_dec import ida_decompile
from .ghidra_dec import ghidra_decompile
from .. import ALL_DECOMPILERS, SAILR_DECOMPILERS
from ..utils import SAILR_DECOMPILATION_RESULTS_DIR, SAILR_COMPILATION_RESULTS_DIR, timeout
from ..utils.sailr_target import SAILRTarget
from ..utils.binary_debug_info import dump_dwarf_addr_to_line_map, dump_pdb_addr_to_line_map
from ..utils.compile import strip_file
from pathlib import Path
from multiprocessing import Pool
from multiprocessing.pool import ApplyResult
from tqdm import tqdm
import logging


l = logging.getLogger(__name__)


def _get_source(source_path, functions=None):
    return source_path


def _get_expanded_source(source_path, functions=None):
    source_path = Path(source_path).absolute()
    source_path = source_path.with_suffix(".i")
    return source_path


def get_decompile_func(dec_name):
    decompilers = {
        "source": _get_source,
        "expanded_source": _get_expanded_source,
        "angr_sailr": angr_decompile,
        "angr_improved_phoenix": angr_improved_phoenix_decompile,
        "angr_comb": angr_comb_decompile,
        "angr_phoenix": angr_phoenix_decompile,
        "angr_dream": angr_dream_decompile,
        "ghidra": ghidra_decompile,
        "ida": ida_decompile
    }

    return decompilers[dec_name]


def _copy_source_file_to_bin(binary_path, source_path, decompiler, copy_linemaps=True, sailr_target: SAILRTarget = None):
    source_path = Path(source_path)
    binary_name = binary_path.with_suffix("").name

    if decompiler == SAILR_DECOMPILERS.SOURCE_EXPANDED_CODE:
        source_path = source_path.with_suffix(".i").absolute()

    out_path = binary_path.parent.joinpath(f"{decompiler}_{binary_name}.c").absolute()
    shutil.copy(source_path, out_path)

    if copy_linemaps:
        # cached results from earlier stuff
        cache_path = source_path.parent.joinpath(f"{decompiler}_{binary_name}.toml")
        if cache_path.exists():
            shutil.copy(cache_path, out_path.parent)

        linemap_path = source_path.parent.joinpath(f"{decompiler}_{binary_name}.linemaps")
        if linemap_path.exists():
            shutil.copy(linemap_path, out_path.parent)
        
        # change the binary path for dumping dwarf, since it will be striped in the sailr decompiled folder
        non_stripped_binary_path = source_path.parent.joinpath(binary_path.name)
        if sailr_target is not None and sailr_target.windows_project:
            l.info(f"Windows project detected, dumping PDB line maps...")
            out = dump_pdb_addr_to_line_map(non_stripped_binary_path)
        else:
            out = dump_dwarf_addr_to_line_map(non_stripped_binary_path)
        shutil.move(out.absolute(), binary_path.parent)

    return out_path


def _decompile(decompiler, binary_path, functions, source_path=None, sailr_target: SAILRTarget = None, symboless=False):
    binary_path = Path(binary_path).absolute()
    if decompiler in (SAILR_DECOMPILERS.SOURCE_CODE, SAILR_DECOMPILERS.SOURCE_EXPANDED_CODE) and source_path:
        if source_path.exists():
            try:
                with timeout(seconds=100):
                    _copy_source_file_to_bin(binary_path, source_path, decompiler, sailr_target=sailr_target)
            except TimeoutError:
                l.info(f"Timout on DWARF collection for {binary_path}")
        return

    dec_func = get_decompile_func(decompiler)
    try:
        output_file = dec_func(binary_path, functions=functions, symboless=symboless)
    except Exception as e:
        l.critical(f"Failed to decompile {binary_path} with {decompiler} because {e}")
        return

    output_path = Path(os.getcwd()).joinpath(output_file)

    new_out_file = binary_path.parent.joinpath(output_path.name).absolute()
    shutil.move(output_path, new_out_file)

    # check for debug info
    if output_path.with_suffix(".linemaps").exists():
        shutil.move(output_path.with_suffix(".linemaps"), new_out_file.with_suffix(".linemaps"))

    # check for pre-metrics
    if output_path.with_suffix(".toml").exists():
        shutil.move(output_path.with_suffix(".toml"), new_out_file.with_suffix(".toml"))

    l.debug(f"Saving file to {new_out_file}")
    return new_out_file


def decompile_file(filepath, sailr_target=None, src_dir=None, functions=None, decompilers=None, cores=1, symboless=False):
    filepath = Path(filepath).absolute()
    source_path = Path(src_dir).joinpath(filepath.with_suffix("").name).with_suffix(".c").absolute() if src_dir else None

    # fill functions only if the decompiler NEEDS it. It's expensive to get all functions (uses angr).
    # this should always skip for only source
    uneeded_decs = {SAILR_DECOMPILERS.SOURCE_CODE, SAILR_DECOMPILERS.GHIDRA, SAILR_DECOMPILERS.IDA}
    if not functions and any([dec not in uneeded_decs for dec in decompilers]):
        functions = get_all_funcs(filepath)

    decompilers = decompilers or ALL_DECOMPILERS
    if not cores or cores == 1:
        # single core it
        for dec in decompilers:
            _decompile(dec, filepath, functions, source_path=source_path, sailr_target=sailr_target, symboless=symboless)
    else:
        # multi-core it
        pool = Pool(processes=len(decompilers))
        pool_results = {}
        for dec in decompilers:
            pool_results[dec] = pool.apply_async(
                _decompile,
                args=(dec, filepath, functions),
                kwds={"sailr_target": sailr_target, "source_path": source_path, "symboless": symboless},
            )
        pool.close()
        pool.join()
        for dec, res in pool_results.items():
            res.get()


#
# Batch functions for speeedup
#

def _prep_and_copy_target(target_package_dir: Path, target: SAILRTarget, clean=False):
    target_package_dir = Path(target_package_dir).absolute()

    # find SAILR compiled object files
    sailr_compile_dir = target_package_dir.joinpath(SAILR_COMPILATION_RESULTS_DIR).absolute()
    if not sailr_compile_dir.exists() and not target.skip_file_copy:
        l.error(f"SAILR Compile dir does not exist at {sailr_compile_dir}")
        return

    # clean and make SAILR decompiled output location
    sailr_decomp_dir = target_package_dir.joinpath(SAILR_DECOMPILATION_RESULTS_DIR).absolute()
    if clean and sailr_decomp_dir.exists():
        shutil.rmtree(sailr_decomp_dir)
        os.mkdir(sailr_decomp_dir)
    elif not sailr_decomp_dir.exists():
        os.mkdir(sailr_decomp_dir)

    # copy all compiled files to decompiled location
    obj_files = sailr_compile_dir.rglob("*.o")
    for obj_file in obj_files:
        shutil.copy(obj_file.absolute(), sailr_decomp_dir)
        new_file = sailr_decomp_dir.joinpath(obj_file.name)
        if not target.windows_project:
            strip_file(new_file)

    obj_file_compile_map = {
        p.expanduser().absolute(): (sailr_compile_dir.absolute(), target) for p in list(Path(sailr_decomp_dir).absolute().rglob("*.o"))
    }
    return obj_file_compile_map


def batch_decompile_and_move_obj_files(target_package_dirs: List[Tuple[Path, SAILRTarget]], decompilers=None, cores=1, clean=False, symboless=False):
    l.info(f"Starting batch decompilation of {len(target_package_dirs)} target directories...")
    # get the mapping for each object file and the source location it should go!
    obj_file_compile_map = {}
    for target_package_dir, target in target_package_dirs:
        obj_file_compile_map.update(_prep_and_copy_target(target_package_dir, target, clean=clean))

    batch_decompile_obj_files(obj_file_compile_map, decompilers=decompilers, cores=cores, symboless=symboless)


def batch_decompile_obj_files(obj_file_compile_map: Dict, decompilers=None, cores=1, symboless=False):
    if not cores or cores == 1:
        for obj_file_path, obj_items in obj_file_compile_map.items():
            obj_src_path, target = obj_items
            decompile_file(obj_file_path, sailr_target=target, decompilers=decompilers, src_dir=obj_src_path, symboless=symboless)
    else:
        with Pool(processes=cores) as pool, tqdm(total=len(obj_file_compile_map), file=sys.stdout) as pbar:
            work = {}
            for obj_file_path, obj_items in obj_file_compile_map.items():
                obj_src_path, target = obj_items
                w = pool.apply_async(
                    decompile_file,
                    args=(obj_file_path,),
                    kwds={"sailr_target": target, "decompilers": decompilers, "src_dir": obj_src_path, "symboless": symboless},
                    callback=lambda _: pbar.update(1)
                )
                job_args = (decompile_file, {"args": obj_file_path, "kwds": {"decompilers": decompilers, "src_dir": obj_src_path}})
                work[w] = job_args

            pool.close()
            for w, job_args in work.items():
                try:
                    w.get(timeout=600)
                except Exception as e:
                    l.critical(f"Job {job_args} failed with exception {e}")
                    pbar.update(1) 
                    continue



from pathlib import Path
import sys
import os

from sailreval.utils.binary_debug_info import read_line_maps, gen_dwarf_addr_to_line_map, dump_dwarf_addr_to_line_map
from sailreval.metrics.ged_to_source import compute_cfg_edit_distance
from sailreval.decompilers.angr_dec import angr_decompile
from sailreval.decompilers.ida_dec import ida_decompile
from sailreval import SAILR_DECOMPILERS

import unittest
from pyjoern import fast_cfgs_from_source
from pyjoern.mapping import correct_source_cfg_addrs
from cfgutils.similarity import graph_edit_distance_core_analysis, ged_upperbound

FILES_PATH = Path(os.path.join(os.path.dirname(os.path.realpath(__file__)), "./cfged/"))
DECOMPILERS = {
    SAILR_DECOMPILERS.SOURCE_CODE: None,
    SAILR_DECOMPILERS.ANGR_SAILR: angr_decompile,
    SAILR_DECOMPILERS.IDA: ida_decompile
}


def extract_cfgs_for_decompiler(source_dir: Path, binary_name, decompiler, function=None):
    source_dir = Path(source_dir).absolute()
    out = fast_cfgs_from_source(source_dir.joinpath(f"{decompiler}_{binary_name}.c"))

    if decompiler == SAILR_DECOMPILERS.SOURCE_CODE:
        linemaps = source_dir.joinpath(f"source_{binary_name}.linemaps")
        out = correct_source_cfg_addrs(out, linemaps)

    return out[function] if function else out


def angr_decompile_and_get_dwarf_info(binary_path: Path, functions=None):
    angr_decompile(binary_path, functions=functions, print_dec=False, dump_line_maps=True)
    dump_dwarf_addr_to_line_map(binary_path)


def cleanup_angr_and_src_files(binary_path: Path, binary_name: str):
    binary_path = Path(binary_path).absolute()
    binary_path.parent.joinpath(f"angr_sailr_{binary_name}.linemaps").unlink()
    binary_path.parent.joinpath(f"angr_sailr_{binary_name}.c").unlink()
    binary_path.parent.joinpath(f"{binary_name}.dwarf.linemaps").unlink()


class TestCFGED(unittest.TestCase):
    def test_dump_linemaps(self):
        # test dwarf dumping
        binary_path = FILES_PATH / "cat/cat.o"
        out = dump_dwarf_addr_to_line_map(binary_path)
        self.assertTrue(out.exists())

        # test angr dumping
        out = angr_decompile(binary_path, dump_line_maps=True)
        self.assertTrue(out.with_suffix(".linemaps").exists())
    
    def test_cfged_is_ged(self):
        binary_path = FILES_PATH/"cat.o"
        filename = binary_path.with_suffix("").name
        decompilers = (SAILR_DECOMPILERS.IDA, SAILR_DECOMPILERS.ANGR_SAILR)
        functions = ["next_line_num", "usage"]
        angr_decompile_and_get_dwarf_info(binary_path, functions=functions)
        for function in functions:
            src_cfg = extract_cfgs_for_decompiler(binary_path.parent, filename, "source", function=function)
            for dec in decompilers:
                dec_cfg = extract_cfgs_for_decompiler(binary_path.parent, filename, dec, function=function)
                cfged_score = compute_cfg_edit_distance(dec_cfg, src_cfg, function, binary_path, dec)
                exact_score = graph_edit_distance_core_analysis(dec_cfg, src_cfg, with_timeout=120)
                print(f"Decompiler: {dec}, Function: {function}, CFGED: {cfged_score}, GED: {exact_score}")
                self.assertEqual(cfged_score, exact_score)

                # special case for IDA
                if dec == SAILR_DECOMPILERS.IDA and function == "next_line_num":
                    self.assertEqual(exact_score, 0)
        cleanup_angr_and_src_files(binary_path, filename)

    def test_cfged_on_large_functions(self):
        """
        Tests that CFGED is less than or equal to the upper bound GED computed on large functions. In this case,
        we use the cat binary which contains at least one large function greater than 80 blocks large.
        We also test the fmt binary.
        """
        targets = {
            "cat": ['cat'],
            "fmt": ['get_paragraph'],
        }
        for target, functions in targets.items():
            binary_path = FILES_PATH/f"{target}.o"
            filename = binary_path.with_suffix("").name
            decompilers = (SAILR_DECOMPILERS.IDA, SAILR_DECOMPILERS.ANGR_SAILR)
            angr_decompile_and_get_dwarf_info(binary_path, functions=functions)
            for function in functions:
                src_cfg = extract_cfgs_for_decompiler(binary_path.parent, filename, "source", function=function)
                for dec in decompilers:
                    dec_cfg = extract_cfgs_for_decompiler(binary_path.parent, filename, dec, function=function)
                    cfged_score = compute_cfg_edit_distance(dec_cfg, src_cfg, function, binary_path, dec)
                    approx_score = ged_upperbound(dec_cfg, src_cfg)
                    print(f"Decompiler: {dec}, Function: {function}, CFGED: {cfged_score}, Upper-GED: {approx_score}")
                    # must be within 10% if it is larger
                    self.assertLessEqual(cfged_score, approx_score*1.10)

            cleanup_angr_and_src_files(binary_path, filename)


if __name__ == "__main__":
    unittest.main(argv=sys.argv)



def count_total_gotos(func_name, client, **kwargs):
    if client is None:
        return None

    if func_name not in client.functions:
        return None

    return client.count_gotos(func_name)



import logging.config
from datetime import datetime

default_config = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {
        "console": {
            "format": "%(levelname)s | %(asctime)s | %(name)-8s | %(message)s"
        },
        "logfile": {
            "format": "%(levelname)s | %(asctime)s | %(name)-8s | %(message)s"
        },
    },

    "handlers": {
        "console": {
            "class": "logging.StreamHandler",
            "level": "DEBUG",
            "formatter": "console",
            "stream": "ext://sys.stdout"
        },
    },
    'loggers': {
        'sailreval': {
            'handlers': ["console"],
            'level': 'INFO',
            'propagate': False
        },
    }
}


class Loggers:
    """
    Logger Manager.
    """
    IN_SCOPE_LOGGERS = ('sailreval')

    def __init__(self):
        self._loggers = {}
        self.load_all_loggers()
        self.profiling_enabled = False

        # disable filelock info logs
        logging.getLogger("filelock").setLevel(logging.WARNING)

        self.config_dict = None
        if default_config is not None:
            self.config_dict = default_config
        if self.config_dict is not None:
            logging.config.dictConfig(self.config_dict)
        self.handler = logging.StreamHandler()
        self.handler.setFormatter(logging.Formatter('%(levelname)-7s | %(asctime)-23s | %(name)-8s | %(message)s'))

    def load_all_loggers(self):
        for name, logger in logging.Logger.manager.loggerDict.items():
            if any(name.startswith(x + '.') or name == x for x in self.IN_SCOPE_LOGGERS):
                self._loggers[name] = logger

    def __getattr__(self, k):
        real_k = k.replace('_', '.')
        if real_k in self._loggers:
            return self._loggers[real_k]
        else:
            raise AttributeError(k)

    def __dir__(self):
        return list(super(Loggers, self).__dir__()) + list(self._loggers.keys())


def is_enabled_for(logger, level):
    if level == 1:
        from .. import loggers
        return loggers.profiling_enabled
    return originalIsEnabledFor(logger, level)


originalIsEnabledFor = logging.Logger.isEnabledFor

# Override isEnabledFor() for Logger class
logging.Logger.isEnabledFor = is_enabled_for



import toml


class SAILRTarget:
    def __init__(self, *args,
                 package_name=None,
                 source_remote=None,
                 download=None,
                 post_download_cmds=None,
                 version=None,
                 package_dir=None,
                 pre_make_cmds=None,
                 make_cmd=None,
                 post_make_cmds=None,
                 source_dir=None,
                 remote_type=None,
                 override_make_vars=None,
                 apply_patch=None,
                 rename_dash_files=False,
                 windows_project=False,
                 skip_compilation=False,
                 subset_compilation=None,
                 skip_file_copy=False,
                 c_compiler=None,
                 cpp_compiler=None,
                 **kwargs
                 ):
        """
        @param args:
        @param package_name: The name associated with this project, used for script settings lookup (this file)
        @param source_remote: The Git or Tar download location
        @param download: True/False, weather this should be downloaded when compiled
        @param post_download_cmds: Shell commands that should be run only after downloading for the first time
        @param version: Version string, used for Git checkout or documentation
        @param package_dir: The name this should be cloned as, and the place entered for compilation (can have subdir)
        @param pre_make_cmds: Commands to run before a make (usually a ./configure)
        @param make_cmd: Usually just 'make'
        @param post_make_cmds: Commands to run after a make (usually a 'rm -rf' of sorts)
        @param source_dir: Directory inside the @package_dir that's used to copy source (and object files) out of
        @param remote_type: git/tar
        @param override_make_vars:
        @param apply_patch: True/False, to apply a patch that is located in the 'targets/patches' folder
        @param rename_dash_files: True/False, Renames files like [expat-foo, expat-bar] to [foo, bar]
        @param kwargs:
        """
        self.package_name = package_name
        self.source_remote = source_remote
        self.download = download or False
        self.post_download_cmds = post_download_cmds
        self.version = version
        self.package_dir = package_dir
        self.pre_make_cmds = pre_make_cmds
        self.make_cmd = make_cmd
        self.post_make_cmds = post_make_cmds
        self.source_dir = source_dir
        self.remote_type = remote_type
        self.override_make_vars = override_make_vars or False
        self.apply_patch = apply_patch
        self.rename_dash_files = rename_dash_files
        self.windows_project = windows_project
        self.skip_compilation = skip_compilation
        self.subset_compilation = subset_compilation
        self.skip_file_copy = skip_file_copy
        self.c_compiler = c_compiler
        self.cpp_compiler = cpp_compiler

    @classmethod
    def from_toml(cls, toml_path):
        with open(toml_path, "r") as fp:
            data = toml.load(fp)

        return cls(**data)



import json
import os
import subprocess
import shutil
from pathlib import Path
import logging
import hashlib

import toml

l = logging.getLogger(__name__)

DUMP_CODE = """
from ghidra.app.decompiler import DecompInterface
from ghidra.program.model.pcode import PcodeOp
from ghidra.app.decompiler import PrettyPrinter

import traceback
import re
import os
import json
import time
from collections import defaultdict, Counter


def collect_call_counts(decomp_func):
    called_functions = list()
    if decomp_func is not None:
        # Iterate over the PcodeOp objects in the decompiled code
        for op in decomp_func.getHighFunction().pcodeOps:
            if op.getOpcode() == PcodeOp.CALL:
                called_func_addr = op.inputs[0].getAddress()
                called_func = currentProgram.getFunctionManager().getFunctionContaining(called_func_addr)
                if called_func is not None:
                    called_functions.append(called_func.name)
    
    out = dict(Counter(called_functions))
    if not out:
        out = {"__empty__": 0}
    
    return out

def collect_countable_metrics(func_text):
    gotos = re.findall("goto [a-zA-Z0-9_]*;", func_text, re.DOTALL)
    goto_cnt = len(gotos)
    and_cnt = func_text.count("&&")
    or_cnt = func_text.count("||")

    counts = {
        "loc": len(func_text.split("\\n")),
        "gotos": goto_cnt,
        "bools": {
            "ands": and_cnt,
            "ors": or_cnt
        }
    }
    return counts

def collect_line_maps(decomp_func):
    linenum_to_addr = defaultdict(set)
    function = decomp_func.function
    linenum_to_addr[1].add(function.entryPoint.offset)
    pp = PrettyPrinter(function, decomp_func.getCCodeMarkup())
    for line in pp.getLines():
        ln = line.getLineNumber()
        for i in range(line.getNumTokens()):
            min_addr = line.getToken(i).getMinAddress()
            if min_addr is None:
                continue

            linenum_to_addr[ln].add(min_addr.offset)
            max_addr = line.getToken(i).getMaxAddress()
            if max_addr is not None:
                linenum_to_addr[ln].add(max_addr.offset)

    list_linenum_to_addr = {
        str(k): list(v) for k,v in dict(linenum_to_addr).items()
    }

    return list_linenum_to_addr


prog_name = currentProgram.name
if "." in prog_name:
    prog_name = prog_name.split(".")[0]
prog_parent_path = os.path.dirname(currentProgram.executablePath)
if prog_parent_path == "/":
    prog_parent_path = "./"
c_output_path = prog_parent_path + "/ghidra_" + prog_name + ".c"
metric_output_path = prog_parent_path + "/ghidra_" + prog_name + ".toml"
linemaps_output_path = prog_parent_path + "/ghidra_" + prog_name + ".linemaps"

output_text = ""
counted_metrics = {}
linemaps = {}
for f in currentProgram.getFunctionManager().getFunctions(True):
    starttime = time.time()
    try:
        di = DecompInterface()
        di.openProgram(currentProgram)
        dec_func = di.decompileFunction(f, 0, None)
        func_text = dec_func.getDecompiledFunction().getC()
    except:
        continue

    output_text += func_text
    output_text += "\\n"
    func_name = str(f.name) + "__SAILR__" + hex(f.entryPoint.offset)[:-1]

    try:
        counted_metrics[func_name] = collect_countable_metrics(func_text)
    except Exception as e:
        continue

    try:
        counted_metrics[func_name].update({"func_calls": collect_call_counts(dec_func)})
    except Exception as e:
        continue

    try:
        counted_metrics[func_name]["dec_time"] = time.time() - starttime
        linemaps[func_name] = collect_line_maps(dec_func)
    except Exception as e:
        continue

with open(c_output_path, "w") as fp:
    fp.write(output_text)

with open(metric_output_path, "w") as fp:
    json.dump(counted_metrics, fp)

with open(linemaps_output_path, "w") as fp:
    json.dump(linemaps, fp)
"""


def _generate_scripts(write_path='/tmp/dump.py', binary_name=None):
    post_code_path = Path(write_path).absolute()
    with open(post_code_path, "w") as fp:
        fp.write(DUMP_CODE)

    return post_code_path


def _convert_json_to_toml_and_fix_name(josn_path: Path, symboless=False):
    with open(josn_path, "r") as fp:
        json_data = json.load(fp)

    new_data = {}
    for func_name, data in json_data.items():
        sym_nam, addr = func_name.split("__SAILR__")
        if symboless:
            new_data[addr] = data
        else:
            new_data[sym_nam] = data

    # XXX: the name will already end in .toml
    with open(josn_path, "w") as fp:
        toml.dump(new_data, fp)

def path_hash_str(path: Path):
    return hashlib.md5(bytes(path)).hexdigest() 

def ghidra_decompile(binary_path, functions=None, symboless=False):
    """
    Taken from mdec with modifications
    """
    binary_path = Path(binary_path).absolute()

    bin_name = binary_path.with_suffix("").name
    output_dec_filename = f"ghidra_{bin_name}.c"
    metric_output_filename = f"ghidra_{bin_name}.toml"
    linemaps_output_filename = f"ghidra_{bin_name}.linemaps"
    post_code_path = _generate_scripts(write_path=f"/tmp/ghidra_{path_hash_str(binary_path)}_{bin_name}.py", binary_name=bin_name)

    # something like: /ghidra/support/analyzeHeadless
    ghidra_path = os.getenv("GHIDRA_PATH") or 'analyzeHeadless'
    original_cwd = os.getcwd()
    output_file = None
    try:
        dir_path = os.path.dirname(binary_path)
        os.chdir(dir_path)
        subprocess.run(
            [f'{ghidra_path}', '.', f'temp_project_{bin_name}',
             '-import', os.path.basename(binary_path),
             '-postScript', f'{post_code_path}']
        )

        shutil.move(output_dec_filename, f"{original_cwd}/{output_dec_filename}")
        output_file = f"{original_cwd}/{output_dec_filename}"

        _convert_json_to_toml_and_fix_name(metric_output_filename, symboless=symboless)
        _convert_json_to_toml_and_fix_name(linemaps_output_filename, symboless=symboless)
        shutil.move(metric_output_filename, f"{original_cwd}/{metric_output_filename}")
        shutil.move(linemaps_output_filename, f"{original_cwd}/{linemaps_output_filename}")
        os.unlink(f"./temp_project_{bin_name}.gpr")
        shutil.rmtree(f"./temp_project_{bin_name}.rep")
    except Exception as e:
        l.critical(f"Ghidra decompilation exeception: {e}")
    finally:
        os.chdir(original_cwd)
        os.unlink(post_code_path)

    return output_file



from collections import Counter, OrderedDict

def count_func_calls(func_name, client, **kwargs):
    if client is None:
        return None

    if func_name not in client.functions:
        return None

    func_calls = client.func_calls_in_func(func_name)
    if not func_calls:
        return {"__empty__": 0}

    counter = Counter(func_calls)
    output = OrderedDict(sorted(counter.items()))
    output = {k:v for k,v in output.items()}
    return output





from typing import List


def count_if_bools(func_name, client, **kwargs):
    if client is None:
        return None

    if func_name not in client.functions:
        return None

    condition_codes: List[str] = client.get_control_structure_conditions(func_name)
    if condition_codes is None:
        return None

    count = {"or": 0, "and": 0}
    for code in condition_codes:
        count["or"] += code.count("||")
        count["and"] += code.count("&&")

    return count



import itertools
import os
import sys
from copy import deepcopy
from pathlib import Path
from collections import defaultdict
from typing import List, Dict, Tuple, Set
from multiprocessing import Pool
import logging
import statistics

from sailreval import ALL_DECOMPILERS, SAILR_DECOMPILERS, SAILR_METRICS, ALL_METRICS
from sailreval.utils import load_tomls_by_bin_name, bcolors
from sailreval.utils.sailr_target import SAILRTarget
from sailreval.utils.compile import DEFAULT_OPTIMIZATION_LEVELS, OPTIMIZATION_LEVELS
from pyjoern import JoernServer, JoernClient

from tqdm import tqdm
import toml
from .visualization import save_data_as_markdown_table, plot_diff_barchar, plot_histogram

l = logging.getLogger(__name__)

DIFFABLE_METRICS = {SAILR_METRICS.BOOLEAN_COUNT, SAILR_METRICS.FUNC_CALLS, SAILR_METRICS.GOTO_COUNT, SAILR_METRICS.BLOCK_COUNT}
#STAT_TYPES = ("sum", "max", "mean", "median")
STAT_TYPES = ("sum", "mean", "median")


def _replace_function_name_with_known_symbol(func_name):
    replace_map = {
        'gettext': 'dcgettext'
    }

    try:
        f = replace_map[func_name]
    except KeyError:
        f = func_name

    return f


def unique_functions_in_set(fullnames: Set, data_by_pkg_name: Dict, decompiler, metrics):
    bad_funcs = set()
    for fullname in fullnames:
        split_name = fullname.split(".")
        bin_name, func_name = split_name[0], ".".join(split_name[1:])
        func_is_bad = False
        for pkg_name, data_by_bin_name in data_by_pkg_name.items():
            bin_data = data_by_bin_name.get(bin_name, None)
            if bin_data is None:
                #print(f"No source data for {pkg_name} on {bin_name}")
                func_is_bad = True
                break

            dec_data = bin_data.get(decompiler, None)
            if dec_data is None:
                #print(f"No decompiler data for {pkg_name} on {bin_name} for {decompiler}")
                func_is_bad = True
                break

            for metric, metric_data in dec_data.items():
                if metric not in metrics:
                    continue

                if func_name not in metric_data.keys():
                    #print(f"No metric data for {pkg_name} on {bin_name}.{func_name} for {metric} ")
                    func_is_bad = True
                    break

            if func_is_bad:
                break

        if func_is_bad:
            bad_funcs.add(fullname)

    unique_funcs = fullnames - bad_funcs
    return unique_funcs


def normalize_measurable_functions(data_by_pkg_name: Dict[str, Dict[str, Dict]], decompilers: List, metrics: List):
    """
    This function will return a list of fullnames (bin_name.func_name) that are common in every metric
    in every metric in every pkg in the data.

    Note: if you are collecting this data with the "SOURCE" decompiler, which is gathered using JOERN, then this data
    will actually already be unique and not need any more parsing!
    """

    # first, collect every possible fullname function in the first package which will act as our base for
    # removing functions that are not common across all packages
    first_pkg = data_by_pkg_name[list(data_by_pkg_name.keys())[0]]
    first_dec = decompilers[0]
    all_funcs = set()
    for bin_name, data_by_decompiler in first_pkg.items():
        for decompiler, data_by_metric in data_by_decompiler.items():
            if decompiler not in decompilers:
                continue

            for metric, data_by_func in data_by_metric.items():
                if metric not in metrics:
                    continue

                for func in data_by_func.keys():
                    all_funcs.add(f"{bin_name}.{func}")

    # while here, grab the functions unique to the source code only
    all_first_dec_funcs = set()
    if first_dec != SAILR_DECOMPILERS.SOURCE_CODE:
        l.warning(f"First decompiler is not source code... the output total functions will be of {first_dec}")

    for bin_name, data_by_decompiler in first_pkg.items():
        if first_dec not in data_by_decompiler:
            l.warning(f"Skipping {bin_name} becaused it lacked {first_dec} data...")
            continue 

        data_by_metric = data_by_decompiler[first_dec]
        for metric, data_by_func in data_by_metric.items():
            if metric not in metrics:
                continue

            for func in data_by_func.keys():
                all_first_dec_funcs.add(f"{bin_name}.{func}")
    unique_funcs_in_base = unique_functions_in_set(all_first_dec_funcs, data_by_pkg_name, first_dec, ["gotos"])

    # second, now that we have a list of all possible functions for analysis, we need to eliminate every
    # function that is not present in every package, for every decompiler, for every metric!
    funcs_not_common = set()
    failed_bin_for_dec = {dec: set() for dec in decompilers}
    for fullname in all_funcs:
        is_not_common = False
        split_name = fullname.split(".")
        bin_name, func_name = split_name[0], ".".join(split_name[1:])

        for pkg_name, data_by_bin_name in data_by_pkg_name.items():
            bin_data = data_by_bin_name.get(bin_name, None)
            if bin_data is None:
                is_not_common = True
                break
            
            for decompiler in decompilers:
                dec_data = bin_data.get(decompiler, None)
                if dec_data is None:
                    if bin_name not in failed_bin_for_dec[decompiler]:
                        l.critical(f"Decompiler {decompiler} failed to decompile the full binary {bin_name}")
                        failed_bin_for_dec[decompiler].add(bin_name) 

                    is_not_common = True
                    break 

                for metric in metrics:
                    data_by_func = dec_data.get(metric, None)
                    if data_by_func is None:
                        is_not_common = True
                        break 
                
                    if func_name not in data_by_func.keys():
                        is_not_common = True
                        break

                if is_not_common:
                    break

            if is_not_common:
                break

        if is_not_common:
            funcs_not_common.add(fullname)

    common_funcs = all_funcs - funcs_not_common
    return common_funcs, unique_funcs_in_base


def find_unique_functions_by_binary(data_by_bin_name: Dict, decompilers: List, metrics):
    metadata = {
        "total_functions_parsed_in_source": 0,
        "total_unique_functions_in_source": 0,
        "total_unique_functions_in_all_metrics": 0
    }

    # XXX: you need to change this sometimes
    exception_funcs = ['main', 'usage']
    skip_metrics = []

    # first, collect a set of seen function symbols that are unique
    seen_source_func_names = set()
    unique_source_func_fullnames = set()
    first_dec = SAILR_DECOMPILERS.SOURCE_CODE if SAILR_DECOMPILERS.SOURCE_CODE in decompilers \
        else decompilers[0]
    for bin_name, bin_data in data_by_bin_name.items():
        source_data = bin_data.get(first_dec, None)
        if source_data is None:
            continue

        # we only will check the first metric for each binary
        for metric, metric_data in source_data.items():
            for func in metric_data.keys():
                metadata["total_functions_parsed_in_source"] += 1

                # don't record seen func names
                if func in seen_source_func_names and func not in exception_funcs:
                    continue

                seen_source_func_names.add(func)
                unique_source_func_fullnames.add(f"{bin_name}.{func}")

    # we now have a list of unique symbols, but now we need to verify that each of these symbols appear
    # as a recorded key for every metric for every decompiler
    metadata["total_unique_functions_in_source"] = len(unique_source_func_fullnames)
    delinquent_funcs = set()
    for fullname in unique_source_func_fullnames:
        split_name = fullname.split(".")
        binary, func = split_name[0], ".".join(split_name[1:])

        bin_data = data_by_bin_name.get(binary, None)
        if bin_data is None:
            delinquent_funcs.add(fullname)
            continue

        if any([dec not in bin_data.keys() for dec in decompilers]):
            delinquent_funcs.add(fullname)
            #print(f"A decompiler is missing in {binary}, skipping {fullname}...")
            continue

        is_delinquent = False
        for dec, dec_data in bin_data.items():
            if dec not in decompilers:
                continue

            for metric, metric_data in dec_data.items():
                if metric in skip_metrics:
                    continue

                if metric not in metrics:
                    continue

                if func not in metric_data.keys():
                    #print(f"{func} not in {metric} for {binary} on {dec}")
                    is_delinquent = True
                    break

            # leave early if a delinquent is found
            if is_delinquent:
                break

        if is_delinquent:
            delinquent_funcs.add(fullname)

    # the list of unique functions found in all decompiler outputs is the difference of the two sets
    unique_funcs_by_binary = unique_source_func_fullnames - delinquent_funcs

    # finally, remove every single function that may not exist in the final
    remove_fullnames = set()
    for fullname in unique_funcs_by_binary:
        split_name = fullname.split(".")
        binary, func_name = split_name[0], ".".join(split_name[1:])
        bin_data = data_by_bin_name[binary]

        for decompiler in decompilers:
            dec_data = bin_data.get(decompiler, None)
            if decompiler is None:
                remove_fullnames.add(fullname)
                break

            should_break = True
            for metric in metrics:
                met_data = dec_data.get(metric, None)
                if met_data is None:
                    remove_fullnames.add(fullname)
                    break

                met_val = met_data.get(func_name, None)
                if met_val is None:
                    remove_fullnames.add(fullname)
                    break
            else:
                should_break = False

            if should_break:
                break
        else:
            pass

    unique_funcs_by_binary = unique_funcs_by_binary - remove_fullnames
    metadata["total_unique_functions_in_all_metrics"] = len(unique_funcs_by_binary)
    return unique_funcs_by_binary, metadata


def filter_functions_by_block_size(fullname_funcs, data_by_pkg, decompilers, max_size=None, min_size=None):
    if max_size is None and min_size is None:
        return fullname_funcs

    bad_functions = set()
    for fullname in fullname_funcs:
        split_name = fullname.split(".")
        binary, func_name = split_name[0], ".".join(split_name[1:])

        for pkg, data_by_bin_name in data_by_pkg.items():
            bin_data = data_by_bin_name[binary]

            for decompiler in decompilers:
                dec_data = bin_data[decompiler]
                func_size = dec_data.get(SAILR_METRICS.BLOCK_COUNT, {}).get(func_name, None)
                if func_size is None or \
                        (min_size is not None and func_size < min_size) or \
                        (max_size is not None and func_size > max_size):
                    bad_functions.add(fullname)
                    break

    return fullname_funcs - bad_functions


def summarize_directory(
    data_by_binary: Path, unique_functions_by_binary, funcs_in_src, decompilers: List, metrics: List, unique_funcs=None,
    min_func_size=None, max_func_size=None, show_stats=False, show_diffs=False
):
    from sailreval.metrics import POST_METRICS
    summary = {dec: defaultdict(int) for dec in decompilers}
    summary["metadata"] = {}
    summary["metadata"]["total_unique_functions_in_src"] = len(funcs_in_src)
    summary["metadata"]["total_unique_functions_in_all_metrics"] = len(unique_functions_by_binary)
    listed_metrics = {dec: defaultdict(list) for dec in decompilers}
    data_by_metric = {dec: defaultdict(dict) for dec in decompilers}
    non_post_metrics = [m for m in metrics if m not in POST_METRICS]
    for fullname in unique_functions_by_binary:
        split_name = fullname.split(".")
        binary, func_name = split_name[0], ".".join(split_name[1:])
        bin_data = data_by_binary[binary]
        
        # filter out too-high CFGED functions
        #skip = False
        #if SAILR_METRICS.CFGED in metrics:
        #    vals_for_dec_set = set()
        #    for decompiler in decompilers:
        #        if decompiler == SAILR_DECOMPILERS.SOURCE_CODE:
        #            continue
        #        met_val = bin_data[decompiler][SAILR_METRICS.CFGED][func_name]
        #        vals_for_dec_set.add(int(float(met_val)))
        #    if min(vals_for_dec_set) >= 800:
        #        skip = True
        #        skip = False

        #if skip:
        #    continue

        # key errors SHOULD NEVER happen now, if it does, we should crash
        for decompiler in decompilers:
            dec_data = bin_data[decompiler]
            for metric in non_post_metrics:
                was_set = False
                met_val = dec_data[metric][func_name]
                if isinstance(met_val, dict):
                    int_val = sum(list(met_val.values()))
                    data_by_metric[decompiler][metric][fullname] = met_val
                    was_set = True
                elif isinstance(met_val, str):
                    # this is likely CFGED or GED
                    int_val = int(float(met_val))
                elif isinstance(met_val, int):
                    int_val = met_val
                elif isinstance(met_val, float):
                    int_val = met_val
                    if metric != SAILR_METRICS.DEC_TIME:
                        int_val = int(met_val)
                else:
                    raise Exception(f"This should not be possible. Bad metric: {metric}")

                summary[decompiler][metric] += int_val
                listed_metrics[decompiler][metric].append(int_val)
                if not was_set:
                    data_by_metric[decompiler][metric][fullname] = int_val

    #
    # Diffs!
    #

    if show_diffs:
        for metric in metrics:
            if metric not in DIFFABLE_METRICS:
                continue

            for dec in decompilers:
                listed_metrics[dec][f"diff_{metric}"] = _compute_metric_diff_to_src(metric, dec, data_by_metric)

    #
    # Post Metrics (metrics computed after first analysis)
    #

    from sailreval.metrics.post_metrics import run_post_metrics

    listed_metrics = run_post_metrics([m for m in metrics if m in POST_METRICS], listed_metrics)

    #
    # Mode, Median, and Max.
    #

    for dec in decompilers:
        for metric in metrics:
            if metric not in metrics:
                continue

            metric_names = [metric] if metric not in DIFFABLE_METRICS else [metric, f"diff_{metric}"]
            for metric_name in metric_names:
                metric_list = listed_metrics[dec].get(metric_name, None)
                if not metric_list:
                    if ("diff" in metric_name and show_diffs) or "diff" not in metric_name:
                        l.warning(f"Metric list {metric_name} is empty, skipping statistics on it...")
                    continue

                # rename it to be sum to be clearer
                if metric_name in summary[dec]:
                    del summary[dec][metric_name]
                summary[dec][f"{metric_name}_sum"] = round(sum(metric_list), ndigits=2)

                if show_stats:
                    summary[dec][f"{metric_name}_mean"] = round(statistics.mean(metric_list), ndigits=2)
                    summary[dec][f"{metric_name}_median"] = round(statistics.median(metric_list), ndigits=2)
                    summary[dec][f"{metric_name}_max"] = round(max(metric_list), ndigits=2)

    print(f"{bcolors.BOLD}{bcolors.OKGREEN}======= START SUMMARY ======={bcolors.ENDC}")
    print(toml.dumps(summary))
    print(f"{bcolors.BOLD}{bcolors.OKGREEN}======= END SUMMARY ======={bcolors.ENDC}")

    # XXX: disabled for now
    # save data as various formats
    #decs_by_metric = defaultdict(dict)
    #listed_metrics = dict(listed_metrics)
    #for dec, metric_datas in listed_metrics.items():
    #    for metric, mdata in metric_datas.items():
    #        decs_by_metric[metric][dec] = mdata

    #if SAILR_DECOMPILERS.ANGR_SAILR in decompilers and SAILR_DECOMPILERS.IDA in decompilers and \
    #        SAILR_METRICS.CFGED in metrics:
    #    l.info(f"Plotting difference...")
    #    plot_diff_barchar(decs_by_metric, "cfged", ["angr_sailr", "ida"])

    return summary, listed_metrics


def _veryify_toml_dir_paths(toml_dir_paths, decompilers, metrics):
    # sanity check metrics and decompilers!
    if not decompilers or any([dec not in ALL_DECOMPILERS for dec in decompilers]):
        raise Exception(f"Provided decompilers: {decompilers} not in supported list of decompilers: {ALL_DECOMPILERS}")
    if not metrics or any([m not in ALL_METRICS for m in metrics]):
        raise Exception(f"Provided metrics: {metrics} not in supported list of metrics: {ALL_METRICS}")

    # sanity check toml dir paths
    for toml_dir_path in toml_dir_paths:
        toml_dir_path = Path(toml_dir_path).absolute()
        if not toml_dir_path.exists():
            raise Exception(f"Provided toml directory: {toml_dir_path} does not exist!")


def summarize_data_tomls(
    toml_dir_paths: List[Path], decompilers: List, metrics: List, unique_funcs=None, min_func_size=None,
    max_func_size=None, show_stats=False, show_diffs=False, report_func_order=False
):
    from sailreval.metrics import POST_METRICS
    _veryify_toml_dir_paths(toml_dir_paths, decompilers, [m for m in metrics if m not in POST_METRICS])
    
    # load every package, which can be one, into a dict
    data_by_pkg = {}
    for toml_dir_path in toml_dir_paths:
        for opt_level in OPTIMIZATION_LEVELS:
            if opt_level in str(toml_dir_path):
                break
        else:
            raise ValueError("This path has no optimization level!")
        
        save_name = toml_dir_path.name
        if toml_dir_path.name == "sailr_measured":
            save_name = toml_dir_path.parent.name

        short_name = f"{opt_level}/{save_name}"
        if short_name in data_by_pkg:
            short_name = f"{toml_dir_path.parent.parent.parent.name}/{opt_level}/{save_name}"
        
        data_by_pkg[short_name] = load_tomls_by_bin_name(toml_dir_path)

    # normalize data by assuring each function exists in each: pkg, decompiler, metric, func!
    print(f"Normalizing data for: {data_by_pkg.keys()}")
    unique_functions_by_binary, funcs_in_src = normalize_measurable_functions(
        data_by_pkg, decompilers, [m for m in metrics if m not in POST_METRICS]
    )

    # filter all functions by block size if necessary
    if min_func_size is not None or max_func_size is not None:
        unique_functions_by_binary = filter_functions_by_block_size(
            unique_functions_by_binary, data_by_pkg, decompilers, max_size=max_func_size, min_size=min_func_size
        )

    summary = {}
    listed_metrics = {}
    unique_functions_by_binary = sorted(list(unique_functions_by_binary))
    for pkg, data_by_binary in data_by_pkg.items():
        l.info(f"Summarizing... {pkg}")
        _summary, _listed_metrics = summarize_directory(
            data_by_binary, unique_functions_by_binary, funcs_in_src, decompilers, metrics, unique_functions_by_binary,
            show_stats=show_stats, show_diffs=show_diffs,
        )
        summary[pkg] = _summary
        listed_metrics[pkg] = _listed_metrics
        if report_func_order:
            listed_metrics[pkg]["func_order"] = [f for f in unique_functions_by_binary]


    l.info("Saving markdown table...")
    print(save_data_as_markdown_table(summary, show_stats=show_stats))

    return summary, listed_metrics


def summarize_sailr_targets(
    sailr_targets: List[SAILRTarget], decompilers: List, metrics: List, opt_levels=DEFAULT_OPTIMIZATION_LEVELS,
    unique_funcs=None, min_func_size=None, max_func_size=None, show_stats=False, show_diffs=False,
):
    toml_dirs_by_pkg = defaultdict(list)
    for sailr_target in sailr_targets:
        for opt_level in opt_levels:
            toml_dirs_by_pkg[sailr_target.package_name].append(Path(f"./results/{opt_level}/{sailr_target.package_dir}/sailr_measured"))

    # sanity check
    all_toml_dir_paths = list(itertools.chain.from_iterable(toml_dirs_by_pkg.values()))
    _veryify_toml_dir_paths(all_toml_dir_paths, decompilers, metrics)

    summaries_by_opt = defaultdict(dict)
    listed_metrics_by_opt = defaultdict(dict)
    full_summary_doc = ""
    for pkg_name, toml_dirs in toml_dirs_by_pkg.items():
        _summary, _listed_metrics = summarize_data_tomls(
            toml_dirs, decompilers, metrics, unique_funcs=unique_funcs, min_func_size=min_func_size,
            max_func_size=max_func_size,
        )
        full_summary_doc += save_data_as_markdown_table(deepcopy(_summary), show_stats=show_stats)
        #pkg_name = list(_summary.keys())[0]
        for name_with_opt in _summary:
            for opt_level in opt_levels:
                if opt_level in name_with_opt:
                    break
            else:
                raise ValueError("Opt not in package name")

            summaries_by_opt[opt_level][pkg_name] = _summary[name_with_opt]
            listed_metrics_by_opt[opt_level][pkg_name] = _listed_metrics[name_with_opt]

    with open("full_summary.md", "w") as fp:
        fp.write(full_summary_doc)

    big_summary = {}
    for opt in opt_levels:
        big_summary[opt] = {dec: defaultdict(int) for dec in decompilers + ['metadata']}
    for opt, pkg_data in summaries_by_opt.items():
        for pkg_name, summary in pkg_data.items():
            for dec, data in summary.items():
                for m, mdata in data.items():
                    big_summary[opt][dec][m] += mdata
    
    #
    # Big Stats!
    #

    listed_metrics = {}
    for opt in opt_levels:
        listed_metrics[opt] = {dec: defaultdict(list) for dec in decompilers}
    for opt, pkg_data in listed_metrics_by_opt.items():
        for pkg_name, lmetrics in pkg_data.items():
            for dec in decompilers:
                for metric in metrics:
                    listed_metrics[opt][dec][metric] += lmetrics[dec][metric]

        for dec in decompilers:
            for metric in metrics:
                if metric not in metrics:
                    continue

                metric_names = [metric] if metric not in DIFFABLE_METRICS else [metric, f"diff_{metric}"]
                for metric_name in metric_names:
                    metric_list = listed_metrics[opt][dec].get(metric_name, None)
                    if not metric_list:
                        if ("diff" in metric_name and show_diffs) or "diff" not in metric_name:
                            l.warning(f"Metric list {metric_name} is empty, skipping statistics on it...")
                        continue

                    # rename it to be sum to be clearer
                    if metric_name in big_summary[opt][dec]:
                        del big_summary[opt][dec][metric_name]
                    big_summary[opt][dec][f"{metric_name}_sum"] = round(sum(metric_list), ndigits=2)

                    if show_stats:
                        big_summary[opt][dec][f"{metric_name}_mean"] = round(statistics.mean(metric_list), ndigits=2)
                        big_summary[opt][dec][f"{metric_name}_median"] = round(statistics.median(metric_list), ndigits=2)
                        big_summary[opt][dec][f"{metric_name}_max"] = round(max(metric_list), ndigits=2)


    first_opt = opt_levels[0]
    
    #fixed_summary = {}
    #for opt in opt_lvels:
    #    fixed_summary = {dec: {} for dec in decompilers + ['metadata']}
    #for opt, pkg_data in big_summary.items():
    #    for dec, dec_data in pkg_data.items():
    #        for met, met_data in dec_data.items():
    #            fixed_summary[opt][dec][met] = met_data


    #final_summary = {f"{len(summaries_by_opt[first_opt].keys())} Packages across {len(opt_levels)} opt levels": fixed_summary}
    print(save_data_as_markdown_table(big_summary, show_stats=show_stats))
    return big_summary, listed_metrics_by_opt

#
# Diff Metric Computation
#


def _compute_metric_diff_to_src(metric, dec, data_by_metric):
    dec_metric_data = data_by_metric[dec][metric]
    src_metric_data = data_by_metric[SAILR_DECOMPILERS.SOURCE_CODE][metric]
    diff_list = []

    # special case for function diffing
    if metric == SAILR_METRICS.FUNC_CALLS:
        return _compute_dup_diff(dec, data_by_metric)

    for func, data in src_metric_data.items():
        if isinstance(data, dict):
            src_val = sum(list(data.values()))
        else:
            src_val = data

        dec_val = dec_metric_data.get(func, 0)
        if isinstance(dec_val, dict):
            dec_val = sum(list(dec_val.values()))

        diff_list.append(abs(src_val - dec_val))

    return diff_list


def _compute_dup_diff(dec, data_by_metric):
    dec_metric_data = data_by_metric[dec][SAILR_METRICS.FUNC_CALLS]
    source_metric_data = data_by_metric[SAILR_DECOMPILERS.SOURCE_CODE][SAILR_METRICS.FUNC_CALLS]
    diff_list = []

    for top_func, data in source_metric_data.items():
        for func_call, val in data.items():
            dec_val = dec_metric_data[top_func].get(func_call, 0)
            diff_list.append(abs(val - dec_val))

    return diff_list

#
# Cache to Results tomls
#


def find_decompiler_cache_files(decompilation_cache: Path) -> List[Tuple[Path, str, str]]:
    """
    Output: [(toml_path, decompiler, binary_name)]
    """
    decompilation_cache = Path(decompilation_cache)
    partial_tomls = [p.absolute() for p in decompilation_cache.glob("*.toml")]
    decompiler_cache_files = []
    for toml_path in partial_tomls:
        for dec_name in ALL_DECOMPILERS:
            if toml_path.name.startswith(dec_name + "_"):
                split_data = toml_path.name.split(dec_name + "_")
                if len(split_data) > 1:
                    bin_name = split_data[-1].split(".toml")[0]
                    break
        else:
            # decompiler name was not found for this target
            continue

        decompiler_cache_files.append((toml_path, dec_name, bin_name))

    return decompiler_cache_files


def collect_and_store_cache_data(measurement_dir: Path, decompilation_cache: Path):
    measurement_dir = Path(measurement_dir)
    toml_paths = [p.absolute() for p in measurement_dir.glob("*.toml")]
    prev_meas_data_by_bin = defaultdict(dict)
    dec_cache_data_by_binary = decompilation_cache_to_results_format(decompilation_cache)
    for toml_path in toml_paths:
        with open(toml_path, "r") as fp:
            toml_data = toml.load(fp)

        bin_name = toml_data['binary']
        for dec in ALL_DECOMPILERS:
            if dec in toml_data:
                prev_meas_data_by_bin[bin_name][dec] = toml_data[dec]

    # base all initial cache on decompiler directly caches measurements
    final_data = dec_cache_data_by_binary.copy()
    for bin_name, data_by_decompiler in dec_cache_data_by_binary.items():
        for dec in data_by_decompiler.keys():
            meas_metric_data = prev_meas_data_by_bin.get(bin_name, {}).get(dec, None)
            if not meas_metric_data:
                continue

            for metric, mdata in meas_metric_data.items():
                if mdata:
                    final_data[bin_name][dec][metric] = mdata

    # final data is computed, now we save it
    cache_dir = measurement_dir / "cache"
    cache_dir = cache_dir.absolute()
    if not cache_dir.exists():
        cache_dir.mkdir()

    for bin_name, data_by_decompiler in final_data.items():
        data_by_decompiler['binary'] = bin_name
        with open(cache_dir / f"{bin_name}.toml", "w") as fp:
            toml.dump(data_by_decompiler, fp)

    return cache_dir


def decompilation_cache_to_results_format(decompilation_cache: Path, delete_partials=False):
    tdata_by_binary = defaultdict(dict)
    for toml_path, dec_name, bin_name in find_decompiler_cache_files(decompilation_cache):
        with open(toml_path, "r") as fp:
            if dec_name == "ghidra":
                try:
                    toml_data = toml.load(fp)  
                except Exception as e:
                    l.critical(f"Failed to load a ghidra file {toml_path} because {e}")
                    continue
            else:
                toml_data = toml.load(fp)

        # XXX: add source different from other decompilers, since it's not flipped
        if dec_name == SAILR_DECOMPILERS.SOURCE_CODE:
            tdata_by_binary[bin_name][SAILR_DECOMPILERS.SOURCE_CODE] = toml_data
            continue

        data_by_metric = defaultdict(dict)
        for func, metrics in toml_data.items():
            for metric, mdata in metrics.items():
                data_by_metric[metric][func] = mdata

        tdata_by_binary[bin_name][dec_name] = data_by_metric

    for bin_name, toml_data in tdata_by_binary.items():
        toml_data['binary'] = bin_name
        with open(decompilation_cache / f"{bin_name}.toml", "w") as fp:
            toml.dump(toml_data, fp)

    return tdata_by_binary

#
# Metric bigger counting
#

def find_functions_with_bigger_metric(data, decompilers, metric, base_is_zero=False, diff_amt=0):
    def _format_number(num):
        if isinstance(num, str):
            return int(float(num))

        return num

    #
    # do it!
    #

    old_dec = decompilers[0]
    new_dec = decompilers[1]
    functions = set()

    for bin_name, bin_data in data.items():
        old_goto = bin_data.get(old_dec, {}).get(metric, None)
        new_goto = bin_data.get(new_dec, {}).get(metric, None)
        if new_goto is None or old_goto is None:
            continue

        for func, old_score in old_goto.items():
            if func not in new_goto:
                continue

            old_score = _format_number(old_score)
            new_score = _format_number(new_goto[func])
            if (base_is_zero and old_score == 0) or not base_is_zero:
                if new_score > old_score and new_score - old_score >= diff_amt:
                    #print(f"{bin_name}.{func}: {new_dec}:{new_score} > {old_dec}:{old_score}")
                    functions.add(f"{bin_name}.{func}")

    return functions


#
# Switch counting
#

def _find_functions_with_switches(source_path: Path, port):
    source_name = source_path.name.split(".c")[0]

    with JoernServer(port=port):
        client = JoernClient(source_path, port=port)
        functions = client.functions_with_switches()

    return [f"{source_name}.{func}" for func in functions]


def find_functions_with_switches(source_paths: List[Path], cores=12):
    with Pool(processes=cores) as pool, tqdm(total=len(source_paths)) as pbar:
        work = []

        for i, src_path in enumerate(source_paths):
            if not src_path.exists():
                pbar.update(1)
                continue

            work.append(pool.apply_async(
                _find_functions_with_switches,
                args=(src_path, 9000+i),
                callback=lambda _: pbar.update(1)
            ))


        pool.close()
        result = []
        for w in work:
            try:
                funcs = w.get(timeout=240)
            except:
                pbar.update(1)
                continue

            result += funcs

        with open("results.toml", "w") as fp:
            toml.dump({"switch-functions": result}, fp)



import subprocess
import os
import time
from pathlib import Path
from collections import defaultdict
import shutil

import toml


def ida_decompile(binary_path, functions=None, symboless=False, thirty_two_bit=False):
    max_wait = 50
    binary_path = Path(binary_path).absolute()
    bin_name = binary_path.with_suffix("").name
    ida_path = os.getenv("IDA_PATH") or 'idat64'
    if thirty_two_bit:
        ida_path = ida_path.replace("idat64","idat")

    cmd = f'{ida_path} -A -c -Ohexrays:ida_{bin_name}:ALL -S"{__file__}" {binary_path}'
    subprocess.run([cmd], timeout=300, shell=True)
    start_wait = time.time()
    while f'ida_{bin_name}.c' not in os.listdir("."):
        if time.time() - start_wait > max_wait:
            return None

        time.sleep(5)

    # delete intermediate ida files
    try:
        os.unlink(f"{binary_path.name}.i64")
    except Exception:
        pass

    new_path = Path(f"ida_{bin_name}.c").absolute()
    _clean_ida_file(new_path)

    # try to move linemaps if they were created
    try:
        shutil.move(binary_path.with_suffix(".ida.linemaps"), new_path.with_suffix(".linemaps"))
    except Exception:
        pass

    #with open(binary_path.with_suffix(".ida.maps"), "r") as fp:
    #    bin_func_line_maps = toml.load(fp)
    ## convert keys to int and values to set
    #for func in bin_func_line_maps:
    #    for linenum in list(bin_func_line_maps[func]):
    #        bin_func_line_maps[func][int(linenum)] = set(bin_func_line_maps[func][linenum])
    #        del bin_func_line_maps[func][linenum]

    return new_path


def _clean_ida_file(file_path: Path):
    with open(file_path, "r") as fp:
        dec_data = fp.read()

    # TODO: if we ever do more than x64, this needs to be updated!
    replacement_map = {
        "__int8": "char",
        "__int16": "short",
        "__int32": "int",
        "__int64": "long",
        "__fastcall": "",
        "__noreturn": "",
        "__cdecl": "",
        "::": "_",
    }

    for k, v in replacement_map.items():
        dec_data = dec_data.replace(k, v)

    with open(file_path, "w") as fp:
        fp.write(dec_data)


if __name__ == "__main__":
    # THIS CODE IS ONLY RUN INSIDE IDA
    import idaapi
    import idautils
    import idc
    import re

    from collections import defaultdict, Counter
    
    # TODO: make this configurable outside
    symboless = False

    def collect_call_counts():
        callees = defaultdict(list)
        # loop through all functions
        for function_ea in idautils.Functions():
            f_name = str(idc.get_func_name(function_ea)) if not symboless else hex(function_ea)
            # For each of the incoming references
            for ref_ea in idautils.CodeRefsTo(function_ea, 0):
                # Get the name of the referring function
                caller_name = str(idc.get_func_name(ref_ea)) if not symboless else hex(ref_ea) 
                # Add the current function to the list of functions
                # called by the referring function
                callees[caller_name].append(f_name)

        counts = {
            k: dict(Counter(v)) for k,v in callees.items()
        }
        return counts

    def collect_countable_metrics(cfunc):
        func_text = str(cfunc)
        gotos = re.findall("goto [a-zA-Z0-9_]*;", func_text, re.DOTALL)
        goto_cnt = len(gotos)
        and_cnt = func_text.count("&&")
        or_cnt = func_text.count("||")

        counts = {
            "loc": len(func_text.split("\n")),
            "gotos": goto_cnt,
            "bools": {
                "ands": and_cnt,
                "ors": or_cnt
            }
        }
        return counts

    def dec_line_map(cfunc):
        linenum_to_addr = defaultdict(set)

        # always add the start as line 1
        linenum_to_addr[1].add(cfunc.entry_ea)

        # find all lines 2 - N
        for addr, lines in cfunc.get_eamap().items():
            for line in lines:
                y_holder = idaapi.int_pointer()
                if not cfunc.find_item_coords(line, None, y_holder):
                    continue

                linenum = y_holder.value()
                linenum_to_addr[linenum].add(addr)

        linenum_to_addr = {str(k): v for k, v in linenum_to_addr.items()}
        return linenum_to_addr

    idc.auto_wait()
    func_line_maps = {}
    func_metric_counts = {}
    for func in idautils.Functions():
        starttime = time.time()
        try:
            cfunc = idaapi.decompile(func)
        except Exception:
            continue
        
        if symboless:
            func_name = hex(func) 
        else:
            func_name = str(idaapi.get_func_name(func))


        try:
            func_metric_counts[func_name] = collect_countable_metrics(cfunc)
        except Exception:
            continue

        try:
            func_metric_counts[func_name]["dec_time"] = time.time() - starttime
            func_line_maps[func_name] = dec_line_map(cfunc)
        except:
            continue

    call_counts = collect_call_counts()
    for _func_name in func_metric_counts:
        if _func_name in call_counts:
            counts = call_counts[_func_name]
            if not counts:
                counts = {"__empty__": 0}

            func_metric_counts[_func_name].update({"func_calls": counts})
        else:
            func_metric_counts[_func_name].update({"func_calls": {"__empty__": 0}})

    path = Path(idaapi.get_input_file_path())
    with open(path.with_suffix(".ida.linemaps"), "w") as fp:
        toml.dump(func_line_maps, fp)

    binary_name = path.with_suffix("").name
    parent = path.parent
    with open(parent.joinpath(Path(f"ida_{binary_name}.toml")), "w") as fp:
        toml.dump(func_metric_counts, fp)



from typing import Dict
import networkx as nx


def cyclomatic_complexity(
    func_name, client, source_cfgs: Dict[str, nx.DiGraph] = None, dec_cfgs: Dict[str, nx.DiGraph] = None,
    binary_path=None, decompiler=None, **kwargs
):
    if client is None and not dec_cfgs:
        return None

    if func_name not in dec_cfgs:
        return None

    func_cfg = dec_cfgs[func_name]

    # Cyclomatic Complexity = E - N + 2*C
    # E = edges
    # N = nodes
    # C = strongly connected graphs
    e = len(list(func_cfg.edges))
    n = len(list(func_cfg.nodes))
    cc = e - n + 2
    return cc



import os
import toml
from pathlib import Path

from collections import Counter
from sailreval import ALL_DECOMPILERS, ALL_METRICS

def remove_inconsistant_functions(goto_counter, ged_counter, binary, total_dict, decompilers):
    goto_remove_list = []
    ged_remove_list = []
    valid_decompilers = [x for x in decompilers if binary in total_dict[x]]
    for k, v in goto_counter.items():
        if v != len(valid_decompilers):
            goto_remove_list.append(k)

    for k, v in ged_counter.items():
        if v != len(valid_decompilers):
            ged_remove_list.append(k)
    
    for d in valid_decompilers:
        for func in goto_remove_list:
            try:
                total_dict[d][binary]['gotos'].pop(func)
            except KeyError:
                pass
        for func in ged_remove_list:
            try:
                total_dict[d][binary]['ged'].pop(func)
            except KeyError:
                pass
    return total_dict

def create_summary(total_dict, decompilers):
    summary = {d: dict() for d in decompilers}
    for d, binaries in total_dict.items():
        goto_funcs = 0
        for bin_name, metrics in binaries.items():
            for metric, values in metrics.items():
                if metric == "ged":
                    if metric + "_total" not in summary[d]:
                        summary[d][metric + "_total"] = 0.0
                        summary[d][metric] = dict()
                        summary[d]["total_analyzed_functions_ged"] = 0
                    ged_total = sum([float(x) for x in values.values()])
                    summary[d][metric + "_total"] += ged_total
                    summary[d][metric][bin_name] = ged_total
                    summary[d]["total_analyzed_functions_ged"] += len(values.keys())

                elif metric == "bools":
                    if metric not in summary[d]:
                        summary[d][metric] = {"or": 0, "and": 0}
                    if bin_name not in summary[d][metric]:
                        summary[d][metric][bin_name] = {"or": 0, "and": 0}
                    for func, func_vals in values.items():
                        summary[d][metric]["or"] += func_vals["or"]
                        summary[d][metric]["and"] += func_vals["and"]
                        summary[d][metric][bin_name]["or"] += func_vals["or"]
                        summary[d][metric][bin_name]["and"] += func_vals["and"]

                elif metric in ALL_METRICS:
                    if metric not in summary[d]:
                        summary[d][metric] = 0
                        if metric == "gotos":
                            summary[d][f"total_analyzed_functions_gotos"] = 0
                    if metric == "gotos":
                        goto_funcs += len([k for k, v in values.items() if v > 0])
                        summary[d]["total_analyzed_functions_gotos"] += len(values.keys())
                    
                    summary[d][metric] += sum(values.values())
        summary[d]["gotos_funcs"] = goto_funcs
    return summary

def analyze_dups(total_dict, decompilers):
    bins = list(total_dict["source"].keys())
    decompilers = [x for x in decompilers if x != "source"]
    summary = {d: {"func_calls": {b: None}} for d in decompilers for b in bins}
    for b in bins:
        if "func_calls" not in total_dict["source"][b]:
            continue
        base = {k: Counter(v) for k, v in total_dict["source"][b].pop("func_calls").items()}
        for d in decompilers:
            if b not in total_dict[d]:
                continue
            other = {k: (Counter(v) - base[k]).total() for k, v in total_dict[d][b]["func_calls"].items() if k in base}
            summary[d]["func_calls"][b] = Counter(other).total()
            total_dict[d][b]["func_calls"] = other
    return summary, total_dict


def merged_results_from_directory(dirname, decs=None, remove_errors=True):
    if decs is None:
        decs = ALL_DECOMPILERS
    if "source" not in decs:
        decs.append("source")
    dirname = Path(dirname).absolute()
    targets = [t.absolute() for t in dirname.rglob("*.toml")]
    total = {d: dict() for d in decs}
    for target in targets:
        with open(target, "r") as t:
            data = toml.load(t)
            goto_counter = Counter()
            ged_counter = Counter()
            valid_decompilers = [d for d in decs if d in data]
            for d in valid_decompilers:
                total[d][data['binary']] = data[d]
                if 'gotos' in data[d]:
                    goto_counter.update(data[d]['gotos'].keys())
                if 'ged' in data[d]:
                    ged_counter.update(data[d]['ged'].keys())

            if remove_errors:
                total = remove_inconsistant_functions(goto_counter, ged_counter, data['binary'], total, decs)
    
    dup_summary, total = analyze_dups(total, decs)
    summary = create_summary(total, decs)
    for d in [x for x in decs if x != "source"]:
        try:
            summary[d]["func_calls"] = dup_summary[d]["func_calls"]
        except KeyError:
            continue

    total["summary"] = summary
    return total



import shutil
from pathlib import Path
import os
import logging
import signal

import toml

l = logging.getLogger(__name__)

#
# Consts
#

SAILR_COMPILATION_RESULTS_DIR = "sailr_compiled"
SAILR_DECOMPILATION_RESULTS_DIR = "sailr_decompiled"
SAILR_MEASURE_RESULTS_DIR = "sailr_measured"

#
# printing colors
#

class bcolors:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKCYAN = '\033[96m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'


#
# filesystem helpers
#

class WorkDirContext:
    def __init__(self, path: Path):
        self.path = path
        self.origin = Path(os.getcwd()).absolute()

    def __enter__(self):
        os.chdir(self.path)

    def __exit__(self, exc_type, exc_val, exc_tb):
        os.chdir(self.origin)


def force_mkdir(path: Path):
    path = Path(path).expanduser().absolute()
    if path.exists():
        shutil.rmtree(path, ignore_errors=True)

    try:
        os.mkdir(path)
    except Exception as e:
        l.warning(f"Failed to make directory because {e}")


class timeout:
    def __init__(self, seconds=1, error_message='Timeout'):
        self.seconds = seconds
        self.error_message = error_message

    def handle_timeout(self, signum, frame):
        raise TimeoutError(self.error_message)

    def __enter__(self):
        signal.signal(signal.SIGALRM, self.handle_timeout)
        signal.alarm(self.seconds)

    def __exit__(self, type_, value, traceback):
        signal.alarm(0)


def load_tomls_by_bin_name(dir_path: Path):
    exclusion_names = {"summary.toml", "results.toml"}
    dir_path = Path(dir_path).absolute()
    data_by_bin_name = {}

    for file_path in dir_path.glob("*.toml"):
        if file_path.name in exclusion_names:
            continue

        with open(file_path.absolute(), "r") as fp:
            data = toml.load(fp)

        # XXX: this changes if you change the format
        exclusion_keys = ["binary", "total_time", "timeout"]
        bin_name = data.get("binary", None)

        if bin_name is None:
            continue

        # remove non-data
        for k in exclusion_keys:
            if k in data:
                del data[k]

        data_by_bin_name[bin_name] = data

    return data_by_bin_name



import unittest
from pathlib import Path
import subprocess
import os

import toml

from sailreval import SAILR_DECOMPILERS, SAILR_METRICS

TESTS_DIR = Path(__file__).parent.absolute()
SCRIPTS_DIR = (TESTS_DIR.parent / "scripts").absolute()


def undockerify_script(script_path: Path, output_path: Path) -> Path:
    with open(script_path, "r") as f:
        script = f.read()

    script = script.replace("sudo ./docker-eval.sh", "./eval.py")
    with open(output_path, "w") as f:
        f.write(script)

    os.chmod(str(output_path), os.stat(str(output_path)).st_mode | 0o111)
    return output_path


class TestPipeline(unittest.TestCase):
    def test_eval_script(self):
        """
        Tests every stage in the eval.py script as described in the README.
        """
        opt_level = "O2"
        project = "example_project"
        decompilers = [SAILR_DECOMPILERS.SOURCE_CODE, SAILR_DECOMPILERS.ANGR_SAILR, SAILR_DECOMPILERS.ANGR_DREAM]
        metrics = [SAILR_METRICS.GOTO_COUNT, SAILR_METRICS.BOOLEAN_COUNT, SAILR_METRICS.FUNC_CALLS, SAILR_METRICS.CFGED]

        # take the verification script and replace docker instances (for CI run)
        eval_script_copy = undockerify_script(SCRIPTS_DIR / "verify_pipeline.sh", TESTS_DIR / "verify_pipeline.sh")
        # run the eval script like normal uses would run it (without docker)
        ret = subprocess.run([f"{eval_script_copy}"], shell=True, cwd=TESTS_DIR)
        assert ret.returncode == 0

        # verify the output
        results_file = Path("results.toml")
        assert results_file.exists()
        with open(results_file, "r") as f:
            results = toml.load(f)

        project_results = results[f"{opt_level}/{project}"]
        for dec in decompilers:
            assert dec in project_results
            for metric in metrics:
                # since we are requesting --show-stats in eval, we need to check for the sum instead of normal
                # metric name; normally, this suffix is not needed.
                assert f"{metric}_sum" in project_results[dec]

        # make sure real numbers occurred
        assert project_results[SAILR_DECOMPILERS.SOURCE_CODE][f"{SAILR_METRICS.GOTO_COUNT}_sum"] == 1
        for dec in decompilers:
            assert project_results[dec][f"{SAILR_METRICS.FUNC_CALLS}_sum"] > 0

        # verify we had the same number of functions as in source (8)
        metadata = project_results["metadata"]
        assert metadata["total_unique_functions_in_src"] == 8
        assert metadata["total_unique_functions_in_all_metrics"] == 8

        # cleanup:
        eval_script_copy.unlink()
        results_file.unlink()
        Path("full_summary.md").unlink()
        Path("summary.md").unlink()



import os
import re
import sys

import angr
from sailreval.decompilers.angr_dec import angr_decompile

import unittest

BINARIES_PATH = os.path.join(os.path.dirname(os.path.realpath(__file__)), "./binaries/")


def decompile_function(binary_path, func_name):
    return angr_decompile(binary_path, functions=[func_name], print_dec=True)


def run_many(times):
    def _run_many(function):
        def __run_many(*args, **kwargs):
            for i in range(times):
                result = function(*args, **kwargs)
            return result
        return __run_many
    return _run_many


class TestDeduplicationReverter(unittest.TestCase):
    def test_test1(self):
        """
        Tests the simplest case of deduplication where only a single block is duplicated
        """
        decomp = decompile_function(os.path.join(BINARIES_PATH, "test1"), "foo")
        text: str = decomp

        self.assertEqual(text.count("first print"), 1)
        self.assertEqual(text.count("second print"), 1)
        self.assertEqual(text.count("third print"), 1)
        self.assertEqual(text.count("if"), 2)
        self.assertEqual(text.count("goto"), 0)

    def test_test2(self):
        """
        Tests the case of a multiple blocks being duplicated with overlapping conditions in an
        if-elif-else structure that causes some graph to be duplicated.
        """
        decomp = decompile_function(os.path.join(BINARIES_PATH, "test2"), "foo")
        text: str = decomp

        self.assertEqual(text.count("first print"), 1)
        self.assertLessEqual(text.count("second print"), 1)
        # TODO: thse should be 1, but we give up for now
        self.assertEqual(text.count("third print"), 2)
        self.assertEqual(text.count("fourth print"), 2)
        # TODO: still need to add more fixes to remove the redudant condition graph
        # this is the same problem at the end of test1
        """
        self.assertEqual(text.count("&&"), 2)
        self.assertEqual(text.count("||"), 1)
        self.assertEqual(text.count("if"), 3)
        """

    def _DISABLED_test_true_emit_ancillary_info(self):
        # TODO: this is disabled until we fix gettext having multiple args
        """
        Found in `true` binary, tests deduplication of an entire graph, with simple
        post blocks (no continuing graph)
        decomp = decompile_function(os.path.join(BINARIES_PATH, "coreutils/true.o"), "emit_ancillary_info.constprop.0")
        text: str = decomp.text

        self.assertEqual(text.count("\\n%s online help: <%s>\\n"), 1)
        self.assertEqual(text.count("Full documentation <%s%s>\\n"), 1)
        self.assertEqual(text.count("or available locally via: info '(coreutils) %s%s'\\n"), 1)
        self.assertEqual(text.count("Report any translation bugs to <https://translationproject.org/team/>\\n"), 1)
        """
        assert True

    def test_true_localcharset_get_charset_aliases(self):
        """
        Tests a complicated partial graph where some pre-blocks and post-blocks differ across
        the duplicated graph
        """
        func_addr = 0x404410
        decomp = decompile_function(os.path.join(BINARIES_PATH, "true_localcharset"), func_addr)
        text: str = decomp

        #
        # first fix:
        # removes the duplicated graphs that start with malloc and end in memcpy
        #
        assert text.count("malloc") == 2

        #
        # second fix:
        # removes the duplicated call
        # TODO: fix this, the goto that makes this de-duplication possible is removed right now
        #
        assert text.count("sub_404860") == 1


    def test_sort_zaptemp(self):
        decomp = decompile_function(os.path.join(BINARIES_PATH, "coreutils/sort.o"), "zaptemp")
        text: str = decomp

        self.assertNotIn("{\n}", decomp)
        self.assertEqual(text.count("goto"), 0)
        # If failing, check that the variable name has not changed. Gernerally, this is checking that
        # the last if-stmt exists with only a single variable in its conditions
        null_if_cases = re.findall(r"if \(!*\*\(v4\)\)", decomp)
        self.assertEqual(len(null_if_cases), 1)

    def test_cksum_digest_split_3(self):
        """
        This tests the deduplicator goto-trigger is not too sensitive. In this binary there is duplicate assignment
        that was legit written by the programmer. It so happens to be close to a goto, which used to trigger this opt
        to remove it and cause more gotos. Therefore, this should actually never result in a succesful fixup of the
        assignment.
        """

        decomp = decompile_function(os.path.join(BINARIES_PATH, "coreutils/cksum-digest.o"), "split_3")
        text: str = decomp

        # the `v9` may change, but the `algorithm_bits` should not
        good_assighment = "digest_length = *((&algorithm_bits + 0x4 * v9));"

        assert "{\n}" not in decomp
        assert text.count(good_assighment) == 2


class TestMaximizeSimilarity(unittest.TestCase):
    def test_nohup_main(self):
        """
        The blocks 0x400a75 and 0x400bd9, are closely similar, but are misaligned by 1 instruction that
        should happen before the call. This optimization makes sure that they get optimized to be the largest
        similar amount they can be.
        """
        decomp = decompile_function(os.path.join(BINARIES_PATH, "coreutils/nohup.o"), "main")
        text: str = decomp

        self.assertNotIn("{\n}", decomp)
        self.assertEqual(text.count("isatty(0x2)"), 1)
        self.assertLessEqual(text.count("goto"), 8)


class TestConstPropReverter(unittest.TestCase):
    def test_fmt_main(self):
        """
        This testcase tests 2 things:
        1. that block merger works
        2. that CallArgSimplifier works

        Since the target is a duplicated call where one of the params that was a regsiter
        gets replaced with a variable.

        To better understand why CallArgSimplifier is used see Zions notes for:
        2022-04-14
        2022-04-04
        """
        decomp = decompile_function(os.path.join(BINARIES_PATH, "coreutils/fmt.o"), "main")
        text: str = decomp

        self.assertEqual(text.count("invalid width"), 2)
        self.assertEqual(text.count("xdectoumax"), 2)


class TestReturnDuplicator(unittest.TestCase):
    def test_test_binop(self):
        decomp = decompile_function(os.path.join(BINARIES_PATH, "coreutils/test.o"), "binop")
        text: str = decomp
        
        self.assertNotIn("{\n}", decomp)
        self.assertEqual(text.count("goto"), 0)

    def test_tail_tail_bytes(self):
        decomp = decompile_function(os.path.join(BINARIES_PATH, "coreutils/tail.o"), "tail_bytes.constprop.0")
        text: str = decomp

        self.assertNotIn("{\n}", decomp)
        self.assertEqual(text.count("goto"), 0)

    def test_tsort_search_item(self):
        decomp = decompile_function(os.path.join(BINARIES_PATH, "coreutils/tsort.o"), "search_item")
        text: str = decomp

        # validate not empty output
        self.assertNotIn("{\n}", decomp)
        self.assertGreaterEqual(text.count("assert_fail"), 1)

        self.assertEqual(text.count("goto"), 0)

    def test_dd_iread(self):
        decomp = decompile_function(os.path.join(BINARIES_PATH, "coreutils/dd.o"), "iread")
        text: str = decomp

        # validate not empty output
        self.assertNotIn("{\n}", decomp)
        assert text.count("return") <= 4
        assert text.count("goto") == 0

    def test_stty_recover_mode(self):
        decomp = decompile_function(os.path.join(BINARIES_PATH, "coreutils/stty.o"), "recover_mode")
        text: str = decomp

        # validate not empty output
        self.assertNotIn("{\n}", decomp)
        assert text.count("return") <= 4
        assert text.count("goto") == 0

class TestLoweredSwitchReverter(unittest.TestCase):
    def DISABLED_test_cksum_digest_filename_unescape(self):
        """
        This is disabled because ReturnSimplifier will remove a goto that triggered this. 
        """
        decomp = decompile_function(os.path.join(BINARIES_PATH, "coreutils/cksum-digest.o"), "filename_unescape")
        text: str = decomp

        self.assertEqual(text.count("switch "), 1)
        self.assertEqual(text.count("case 92:"), 1)
        self.assertEqual(text.count("goto"), 1)


class TestCrossJumpReverter(unittest.TestCase):
    def test_chcon_process_file(self):
        decomp = decompile_function(os.path.join(BINARIES_PATH, "coreutils/chcon.o"), "process_file")
        text: str = decomp

        self.assertGreaterEqual(text.count("switch "), 1)
        self.assertLessEqual(text.count("goto"), 1)


if __name__ == "__main__":
    unittest.main(argv=sys.argv)



import logging

from .total_paths import count_total_paths
from .cyclomatic_complexity import cyclomatic_complexity
from .total_gotos import count_total_gotos
from .total_loc import count_total_loc
from .nesting_count import count_nesting_ifs
from .duplication import count_func_calls
from .total_bools import count_if_bools
from .ged_to_source import ged_upperbound_score, cfg_edit_distance, ged_max_score, block_count, ged_exact_score, graph_size
from .post_metrics import norm_cfged, zero_cfged, funcs_w_goto, goto_func_cfged

from sailreval import SAILR_METRICS

"""
Decompilation Metrics:
1. Count Gotos
2. Cyclomatic Complexity

Engineering Metrics:
1. Cyclomatic Complexity: how many control structures do you have?
2. Nesting Count: How complicated is each of those control structures?
3. Total Gotos: How many times did you fail to identify a structure? 
4. Total LoC: How much code you need to read?
5. Irreducible Count: How condensed is that code? <----- ?????
6. Total Paths: How many ways can you analyze the code?

#5 is hard because its unclear in decompilation what is irreducible.

"""

l = logging.getLogger(__name__)

METRICS = {
    # Graph Edit Distance variants
    SAILR_METRICS.GED_UPPERBOUND: ged_upperbound_score,
    SAILR_METRICS.GED_MAX: ged_max_score,
    SAILR_METRICS.GED_EXACT: ged_exact_score,

    # Important Metrics
    SAILR_METRICS.GOTO_COUNT: count_total_gotos,
    SAILR_METRICS.FUNC_CALLS: count_func_calls,
    SAILR_METRICS.CFGED: cfg_edit_distance,

    # Meta Metrics
    SAILR_METRICS.BLOCK_COUNT: block_count,
    SAILR_METRICS.GRAPH_SIZE: graph_size,
    SAILR_METRICS.DEC_TIME: lambda *args, **kwargs: kwargs.get(SAILR_METRICS.DEC_TIME, None),

    # Others
    SAILR_METRICS.PATH_COUNT: count_total_paths,
    SAILR_METRICS.CODE_COMPLEXITY: cyclomatic_complexity,
    SAILR_METRICS.LINES_OF_CODE: count_total_loc,
    SAILR_METRICS.IF_STMT_COUNT: count_nesting_ifs,
    SAILR_METRICS.BOOLEAN_COUNT: count_if_bools,
}

POST_METRICS = {
    SAILR_METRICS.NORM_CFGED: norm_cfged,
    SAILR_METRICS.ZERO_CFGED: zero_cfged,
    SAILR_METRICS.GOTO_FUNC_CFGED: goto_func_cfged,
    SAILR_METRICS.FUNCS_W_GOTO: funcs_w_goto
}


def get_metric_function(metric_name):
    metric_func = METRICS[metric_name]

    def _metric_for_functions(functions, client, **kwargs):
        output = {}
        for func_name in functions:
            try:
                output[func_name] = metric_func(func_name, client, **kwargs)
                timedout = False
            except Exception as e:
                l.critical(f"Failed to run metric function <{metric_func}>({kwargs.get('binary_path', None)}: {func_name}) because {e}")
                output[func_name] = None
                timedout = True

            # XXX: hotfix for get timout to have a max score in timeout event
            if timedout and metric_name in (SAILR_METRICS.GED_UPPERBOUND, SAILR_METRICS.CFGED):
                func_name, source_cfgs, dec_cfgs = kwargs.get("func_name", None), kwargs.get("source_cfgs", None), kwargs.get("dec_cfgs", None)
                if func_name and source_cfgs and dec_cfgs and func_name in source_cfgs and func_name in dec_cfgs:
                    src_cfg = source_cfgs[func_name]
                    dec_cfg = dec_cfgs[func_name]
                    output[func_name] = len(src_cfg.nodes) + len(src_cfg.edges) + len(dec_cfg.nodes) + len(dec_cfg.edges)

        return output

    return _metric_for_functions



import logging
from pathlib import Path
import re
from subprocess import run
from collections import defaultdict

from elftools.elf.elffile import ELFFile
from elftools.dwarf.descriptions import set_global_machine_arch
import toml

l = logging.getLogger(__name__)

#
# DWARF Parsing & Dumping
#

def _lineprog(dwarfinfo, CU):
    line_program = dwarfinfo.line_program_for_CU(CU)
    if not line_program:
        return []
    return line_program.get_entries()


def _start_address(dwarfinfo):
    '''
    starting address to line mapping as in `.debug_line`
    same as readelf's output
    return: addr_to_line and rebased_addr_to_line
    '''
    addr_line = defaultdict(set)
    rebased_addr = dict()

    for CU in dwarfinfo.iter_CUs():
        lpe = _lineprog(dwarfinfo, CU)
        for lp in lpe:
            if not lp.state:
                continue
            # addr: 0x1285 rebased: 0x401285
            # starting address for the line
            addr = lp.state.address
            rebased_addr[(int(0x400000) + addr)] = addr
            line = lp.state.line
            if addr in addr_line:
                addr_line[addr].add(line)
            else:
                addr_line[addr] = {line}

    return addr_line, rebased_addr


def _all_address(dwarfinfo, binpath):
    '''
    starting address to line mapping as in `.debug_line` (readelf's output)
    + possible address in for same line
    '''

    # all instruction address per angr
    all_addrs = _angr_collect_address(binpath)
    # get starting address
    total_addr_line, rebased_addr = _start_address(dwarfinfo)
    # unknown addresses
    # case1: one source line can have multiple instruction addresses which are not shown in `.debug_line`
    # case2: false positive (hopefully not many fp)
    unkn_addr_line = set(all_addrs) - set(rebased_addr)

    for raddr in unkn_addr_line:
        addr = (raddr - int(0x400000))
        for CU in dwarfinfo.iter_CUs():
            lpe = _lineprog(dwarfinfo, CU)
            prev = None
            for lp in lpe:
                if not lp.state:
                    continue
                # prev instruction addr <= current instruction addr < instruction address for next source line
                if prev and prev.address <= addr < lp.state.address:
                    line = prev.line
                    if addr in total_addr_line:
                        total_addr_line[addr].add(line)
                    else:
                        total_addr_line[addr] = {line}
                if lp.state.end_sequence:
                    prev = None
                else:
                    prev = lp.state

    return total_addr_line


def _angr_collect_address(binpath):
    '''
    iterate over all blocks and collect instruction addrs
    '''
    import angr

    all_addrs = []
    project = angr.Project(binpath, auto_load_libs=False, load_debug_info=True)
    cfg = project.analyses.CFG(data_references=True, normalize=True)
    for func in cfg.functions.values():
        if func.is_plt or func.is_simprocedure:
            continue
        if not func.get_block_size(func.addr):
            continue
        for block in func.blocks:
            all_addrs.extend(block.instruction_addrs)

    return all_addrs


def _load_dwarf_info(binpath):
    with open(binpath, 'rb') as f:
        elffile = ELFFile(f)
        if not elffile.has_dwarf_info():
            print("no DWARF")
            return
        set_global_machine_arch(elffile.get_machine_arch())
        dwarfinfo = elffile.get_dwarf_info()
        return dwarfinfo


def gen_dwarf_addr_to_line_map(binary_path: Path, full_search=True):
    binary_path = Path(binary_path).absolute()
    dwarf_info = _load_dwarf_info(binary_path)

    if not full_search:
        total_addr_line, rebased_addr = _start_address(dwarf_info)
        return total_addr_line

    # XXX: uses angr to get all addresses (slow)
    return _all_address(dwarf_info, binary_path)


def dump_dwarf_addr_to_line_map(binary_path: Path, full_search=False):
    addr_to_line = gen_dwarf_addr_to_line_map(binary_path, full_search=full_search)
    output = {
        str(k): v for k, v in addr_to_line.items()
    }

    output_path = binary_path.with_suffix(".dwarf.linemaps")
    with open(output_path, "w") as fp:
        toml.dump(output, fp)

    return output_path

#
# PE Debug Info (PDB) Parsing & Dumping
#


ADDR_LN_MAP_RE = re.compile(r"([0-9]{1,6}|NSI) ([0-9A-F]{8})")
FILE_RE = re.compile(r"([A-Z]:.*\.c)")
MOD_STARTS_RE = re.compile(r"Mod [0-9]{4} \| ")


def _run_pdbutil(filepath: Path):
    text = run(f"llvm-pdbutil dump -l {filepath}".split(" "), capture_output=True).stdout.decode()
    return text


def _extract_line_to_addr_map_for_pdb(filepath: Path):
    filepath = Path(filepath)
    text = _run_pdbutil(filepath)
    start_poss = [
        m.end() for m in MOD_STARTS_RE.finditer(text)
    ]

    end_pos_to_file = {}
    for m in FILE_RE.finditer(text):
        pos = m.start()
        filename = m.group(0)
        for i, start_pos in enumerate(start_poss):
            if i+1 == len(start_poss) and pos >= start_pos:
                end_pos_to_file[start_pos] = filename
                break
            elif pos >= start_pos and pos < start_poss[i+1]:
                end_pos_to_file[start_pos] = filename
                break


    bad_cnt = 0
    for start_pos in start_poss:
        if start_pos not in end_pos_to_file:
            end_pos_to_file[start_pos] = f"ERROR_{bad_cnt}"
            bad_cnt += 1

    end_positions = list(end_pos_to_file.keys())
    # [(start_of_mod, end_of_mod)]
    ranges = []
    for i, end_pos in enumerate(end_positions):
        if i+1 >= len(end_positions):
            r = (end_pos, None)
        else:
            r = (end_pos, end_positions[i+1] - 1)

        ranges.append(r)

    def _find_correct_file(pos):
        for r in ranges:
            start, end = r
            if pos > start and (end is None or pos < end):
                return end_pos_to_file[start]

    mapping = {v: defaultdict(set) for v in end_pos_to_file.values()}
    last_line_number = None
    for m in ADDR_LN_MAP_RE.finditer(text):
        pos = m.start()
        line_number = m.group(1)
        if line_number == "NSI":
            line_number = last_line_number
        else:
            line_number = int(line_number, 10)
            last_line_number = line_number

        addr = int(m.group(2), 16)

        win_filename =  _find_correct_file(pos)
        mapping[win_filename][line_number].add(addr)
    
    dict_mapping = {}
    for win_filename, v in mapping.items():
        if win_filename.startswith("ERROR"):
            continue

        # XXX: a hack for working on non Windows systems
        filename = win_filename.split("\\")[-1]
        if not filename.endswith(".c"):
            l.critical(f"We likely failed to parse name of {win_filename} you should check it...")
        

        dict_mapping[filename] = v
    
    return dict_mapping


def gen_pdb_addr_to_line_map(binary_path: Path):
    source_name = binary_path.with_suffix(".c").name
    line_to_addr_by_src = _extract_line_to_addr_map_for_pdb(binary_path)
    for src, line_to_addr_map in line_to_addr_by_src.items():
        if src == source_name:
            return line_to_addr_map
    else:
        l.critical(f"Failed to find the right source file mappings for {binary_path}")
        return {}


def dump_pdb_addr_to_line_map(binary_path: Path):
    addr_to_line = gen_pdb_addr_to_line_map(binary_path)
    output = {
        str(k): list(v) for k, v in addr_to_line.items()
    }

    output_path = binary_path.with_suffix(".dwarf.linemaps")
    with open(output_path, "w") as fp:
        toml.dump(output, fp)

    return output_path


#
# Extras
#

def read_line_maps(maps_path: Path, key_is_func=True, value_is_set_type=True):
    with open(maps_path, "r") as fp:
        line_maps = toml.load(fp)

    # convert keys to int and values to set
    if key_is_func:
        for func in line_maps:
            for linenum in list(line_maps[func]):
                if value_is_set_type:
                    line_maps[func][int(linenum)] = set(line_maps[func][linenum])
                else:
                    line_maps[func][int(linenum)] = line_maps[func][linenum]
                del line_maps[func][linenum]
    else:
        for addr in list(line_maps.keys()):
            line_maps[int(addr)] = set(line_maps[addr])
            del line_maps[addr]

    return line_maps
